from .cython_substitute import naive_rolling_sum
from .cython_substitute import naive_rolling_min
from .cython_substitute import naive_rolling_max
from .cython_substitute import naive_ema

try:
    from .cython_wrappers import c_rolling_sum as rolling_sum
    from .cython_wrappers import c_rolling_min as rolling_min
    from .cython_wrappers import c_rolling_max as rolling_max
    from .cython_wrappers import c_ema as ema
except ImportError:
    rolling_sum = naive_rolling_sum
    rolling_min = naive_rolling_min
    rolling_max = naive_rolling_max
    ema = naive_ema


__all__ = [
    "rolling_sum",
    "rolling_min",
    "rolling_max",
    "ema",
    "naive_rolling_sum",
    "naive_rolling_min",
    "naive_rolling_max",
    "naive_ema",
]



import cv2

import numpy as np

from typing import *
from PIL import Image


def draw_contours(
    w: int,
    h: int,
    contours: List[np.ndarray],
    *,
    saving_path: str = None,
    color: Tuple[int, int, int] = (0, 255, 0)
) -> Image.Image:
    canvas = np.zeros([h, w, 3]).astype(np.uint8)
    cv2.drawContours(canvas, contours, -1, color, 3)
    im = Image.fromarray(canvas)
    if saving_path is not None:
        im.save(saving_path)
    return im


class Reader:
    def __init__(self, img_path: str, flag: int = cv2.IMREAD_UNCHANGED):
        try:
            self.img_array = cv2.imdecode(np.fromfile(img_path, dtype=np.uint8), flag)
        except:
            self.img_array = None

    @property
    def is_valid(self):
        return self.img_array is not None

    def to_gray(self, *, reverse: bool = False) -> Union[np.ndarray, None]:
        if not self.is_valid:
            return
        if len(self.img_array.shape) == 2:
            gray = self.img_array
        elif self.img_array.shape[-1] == 4:
            gray = self.img_array[..., -1]
        else:
            gray = cv2.cvtColor(self.img_array, cv2.COLOR_RGB2GRAY)
            if reverse:
                gray = 255 - gray
        return gray


__all__ = ["draw_contours", "Reader"]



import math

from ...types import *
from ..bases import SingleNormalizer


@SingleNormalizer.register("cube")
class CubeNormalizer(SingleNormalizer):
    @property
    def bounds(self) -> bounds_type:
        return 0.0, 1.0

    def _init_config(self, **kwargs):
        self._convert_only = kwargs.get("convert_only", True)

    def normalize(self, value: generic_number_type) -> float:
        if self._choices is not None:
            idx = self._choices.index(value)
            return idx / (len(self._choices) - 1)
        if self._convert_only:
            return value
        if self._is_exponential:
            value = math.log(value, self.dist.base)
        return (value - self._lower) / self._diff

    def recover(self, value: float) -> generic_number_type:
        if self._choices is not None:
            idx = int(round(value * (len(self._choices) - 1)))
            idx = max(0, min(len(self._choices) - 1, idx))
            return self._choices[idx]
        if self._convert_only:
            return value
        value = max(0.0, min(1.0, value))
        value = value * self._diff + self._lower
        if self._is_exponential:
            value = math.pow(self.dist.base, value)
        return value


__all__ = ["CubeNormalizer"]



try:
    from .cython_utils import *
except ImportError:
    raise

import numpy as np


def c_rolling_sum(array: np.ndarray, window: int, mean: bool) -> np.ndarray:
    dtype = array.dtype
    return rolling_sum(array.astype(np.float64), window, int(mean)).astype(dtype)


def c_rolling_min(array: np.ndarray, window: int) -> np.ndarray:
    dtype = array.dtype
    array = array.astype(np.float64)
    array[np.isnan(array)] = np.inf
    return rolling_min(array, window).astype(dtype)


def c_rolling_max(array: np.ndarray, window: int) -> np.ndarray:
    dtype = array.dtype
    array = array.astype(np.float64)
    array[np.isnan(array)] = -np.inf
    return rolling_max(array, window).astype(dtype)


def c_ema(array: np.ndarray, window: int) -> np.ndarray:
    dtype = array.dtype
    ratio = 2.0 / (window + 1.0)
    return ema(array.astype(np.float64), ratio).astype(dtype)


__all__ = [
    "c_rolling_sum",
    "c_rolling_min",
    "c_rolling_max",
    "c_ema",
]



import numpy as np

from typing import *

from ...misc import *
from ...ml.param_utils import *

fn_type = Union[Callable[[nested_type], float], None]


class Result(NamedTuple):
    params: nested_type
    score: float


class TargetSpace:
    def __init__(
        self,
        fn: fn_type,
        params: params_type,
        *,
        normalization: Union[str, None],
        normalization_config: Dict[str, Any]
    ):
        self.fn = fn
        self.params_gen = ParamsGenerator(
            params,
            normalize_method=normalization,
            normalize_config=normalization_config,
        )
        self._codes2scores = {}
        self._sorted_keys = self.params_gen.sorted_flattened_keys
        self._tried_flattened_arrays = np.empty([0, self.dim], np.float32)
        self._tried_scores = np.empty([0], np.float32)

    @property
    def dim(self) -> int:
        return self.params_gen.array_dim

    @property
    def num_try(self) -> int:
        return len(self._codes2scores)

    @property
    def tried_flattened_params(self) -> np.ndarray:
        return self._tried_flattened_arrays

    @property
    def tried_scores(self) -> np.ndarray:
        return self._tried_scores

    @property
    def is_empty(self) -> bool:
        return not self._codes2scores

    @property
    def best_result(self) -> Result:
        best_idx = np.argmax(self._tried_scores).item()
        return self._make_result(best_idx)

    @property
    def all_results(self) -> List[Result]:
        return list(map(self._make_result, range(self.num_try)))

    def _make_result(self, idx: int) -> Result:
        array = self._tried_flattened_arrays[idx]
        flattened = self.array2param(array)
        nested = self.params_gen.nest_flattened(flattened)
        return Result(nested, self._tried_scores[idx])

    def param2array(self, param: flattened_type) -> np.ndarray:
        return self.params_gen.flattened2array(param)

    def array2param(self, array: np.ndarray) -> flattened_type:
        return self.params_gen.array2flattened(array)

    def register(self, param: flattened_type, score: float) -> "TargetSpace":
        self._codes2scores[hash_code(str(param))] = score
        param_array = self.param2array(param).reshape(1, -1)
        self._tried_flattened_arrays = np.concatenate(
            [self._tried_flattened_arrays, param_array]
        )
        self._tried_scores = np.concatenate([self._tried_scores, [score]])
        return self

    def probe(self, param: flattened_type) -> "TargetSpace":
        if self.fn is None:
            msg = "fn is not provided, so `probe` method should not be called"
            raise ValueError(msg)
        code = hash_code(str(param))
        score = self._codes2scores.get(code)
        if score is not None:
            return self
        nested = self.params_gen.nest_flattened(param)
        self.register(param, self.fn(nested))
        return self

    def sample(self) -> flattened_type:
        return self.params_gen.flatten_nested(self.params_gen.pop())


__all__ = ["Result", "TargetSpace", "fn_type"]



import math
import random

from abc import *
from typing import *

from .types import *
from .distributions import DistributionBase
from ...misc import prod, Grid


class DataType(metaclass=ABCMeta):
    def __init__(self, distribution: DistributionBase = None, **kwargs):
        self.dist, self.config = distribution, kwargs

    @property
    @abstractmethod
    def num_params(self) -> number_type:
        raise NotImplementedError

    @abstractmethod
    def _transform(self, value) -> generic_number_type:
        raise NotImplementedError

    def __str__(self):
        return f"{type(self).__name__}({self.dist})"

    __repr__ = __str__

    @property
    def lower(self) -> nullable_number_type:
        dist_lower = self.dist.lower
        if dist_lower is None:
            return
        return self._transform(dist_lower)

    @property
    def upper(self) -> nullable_number_type:
        dist_upper = self.dist.upper
        if dist_upper is None:
            return
        return self._transform(dist_upper)

    @property
    def values(self) -> Union[List[generic_number_type], None]:
        dist_values = self.dist.values
        if dist_values is None:
            return
        return list(map(self._transform, dist_values))

    @property
    def bounds(self) -> bounds_type:
        lower, upper = map(self._transform, self.dist.bounds)
        return lower, upper

    @property
    def is_inf(self) -> bool:
        return math.isinf(self.num_params)

    @property
    def distribution_is_inf(self) -> bool:
        return math.isinf(self.dist.num_params)

    def _all(self) -> List[generic_number_type]:
        return self.values

    def pop(self) -> generic_number_type:
        return self._transform(self.dist.pop())

    def all(self) -> List[generic_number_type]:
        if self.is_inf:
            raise ValueError("'all' method could be called iff n_params is finite")
        return self._all()

    def transform(self, value: generic_number_type) -> generic_number_type:
        return self._transform(self.dist.clip(value))


class Any(DataType):
    @property
    def num_params(self) -> number_type:
        return self.dist.num_params

    def _transform(self, value) -> generic_number_type:
        return value


class Int(DataType):
    @property
    def lower(self) -> nullable_number_type:
        dist_lower = self.dist.lower
        if dist_lower is None:
            return
        return int(math.ceil(self.dist.lower))

    @property
    def upper(self) -> nullable_number_type:
        dist_lower = self.dist.upper
        if dist_lower is None:
            return
        return int(math.floor(self.dist.upper))

    @property
    def num_params(self) -> int:
        if self.distribution_is_inf:
            return int(self.upper - self.lower) + 1
        return self.dist.num_params

    def _all(self) -> List[int]:
        if self.distribution_is_inf:
            return list(range(self.lower, self.upper + 1))
        return super()._all()

    def _transform(self, value) -> int:
        return int(round(value + random.random() * 2e-4 - 1e-4))


class Float(DataType):
    @property
    def num_params(self) -> number_type:
        return self.dist.num_params

    def _transform(self, value) -> float:
        return float(value)


class Bool(DataType):
    @property
    def num_params(self) -> int:
        if self.distribution_is_inf:
            return 2
        return len(self._all())

    def _all(self) -> List[bool]:
        return sorted(super()._all())

    def _transform(self, value) -> bool:
        return bool(value)


class String(DataType):
    @property
    def num_params(self) -> number_type:
        return self.dist.num_params

    def _transform(self, value) -> str:
        return str(value)


iterable_data_type = Union[List[DataType], Tuple[DataType, ...]]
iterable_generic_number_type = Union[
    List[generic_number_type], Tuple[generic_number_type, ...]
]


class Iterable:
    def __init__(self, values: iterable_data_type):
        self._values = values
        self.is_list = isinstance(values, list)
        self._constructor = list if self.is_list else tuple

    def __str__(self):
        braces = "[]" if self._constructor is list else "()"
        return f"{braces[0]}{', '.join(map(str, self._values))}{braces[1]}"

    __repr__ = __str__

    def pop(self) -> iterable_generic_number_type:
        return self._constructor(v.pop() for v in self._values)

    def all(self) -> Iterator[generic_number_type]:
        grid = Grid([v.all() for v in self._values])
        for v in grid:
            yield self._constructor(v)

    def transform(
        self,
        value: iterable_generic_number_type,
    ) -> iterable_generic_number_type:
        return self._constructor(v.transform(vv) for v, vv in zip(self._values, value))

    @property
    def values(self) -> iterable_data_type:
        return self._values

    @property
    def bounds(self) -> List[bounds_type]:
        return [v.bounds for v in self._values]

    @property
    def num_params(self) -> number_type:
        num_params = prod(v.num_params for v in self._values)
        if math.isinf(num_params):
            return num_params
        return int(num_params)


__all__ = [
    "DataType",
    "Any",
    "Int",
    "Float",
    "Bool",
    "String",
    "Iterable",
    "iterable_data_type",
    "iterable_generic_number_type",
    "bounds_type",
]



import warnings
import numpy as np

from scipy.stats import norm
from scipy.optimize import minimize
from sklearn.gaussian_process import GaussianProcessRegressor

from .target_space import TargetSpace


class Acquisition:
    def __init__(
        self,
        gp: GaussianProcessRegressor,
        method: str,
        xi: float,
        kappa: float,
        kappa_decay: float,
        kappa_decay_delay: int,
    ):
        self.gp = gp
        self.method = method
        self.xi, self.kappa = xi, kappa
        self.kappa_decay, self.kappa_decay_delay = kappa_decay, kappa_decay_delay
        self._iter_counter = 0

    def __str__(self):
        return f"Acquisition({self.method})"

    __repr__ = __str__

    def update(self) -> "Acquisition":
        self._iter_counter += 1
        if self.kappa_decay < 1 and self._iter_counter > self.kappa_decay_delay:
            self.kappa *= self.kappa_decay
        return self

    def search_max(
        self,
        best_score: float,
        space: TargetSpace,
        num_warmup: int,
        num_iter: int,
    ) -> np.ndarray:
        _array = lambda n: np.array(
            list(map(space.param2array, [space.sample() for _ in range(n)]))
        )
        # warm up
        x_tries = _array(num_warmup)
        scores = self.score(x_tries.astype(np.float32), best_score)
        best_idx = scores.argmax().item()
        best_x, best_score = x_tries[best_idx], scores[best_idx]
        # optimization
        for _ in range(num_iter):
            x_try = _array(1).ravel()
            res = minimize(
                lambda x: -self.score(x.reshape([1, -1]), best_score),
                x_try,
                bounds=space.params_gen.all_bounds,
                method="L-BFGS-B",
            )

            if not res.success:
                continue

            if -res.fun[0] > best_score:
                best_x = res.x
                best_score = -res.fun[0].item()
        return best_x

    def score(self, x: np.ndarray, best_score: float) -> np.ndarray:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            mean, std = self.gp.predict(x, return_std=True)
        return getattr(self, f"_{self.method}")(mean, std, best_score)

    def _ucb(self, mean: np.ndarray, std: np.ndarray, best_score: float) -> np.ndarray:
        return mean + self.kappa * std

    def _ei(self, mean: np.ndarray, std: np.ndarray, best_score: float) -> np.ndarray:
        a = mean - best_score - self.xi
        z = a / std
        return a * norm.cdf(z) + std * norm.pdf(z)

    def _poi(self, mean: np.ndarray, std: np.ndarray, best_score: float) -> np.ndarray:
        return norm.cdf((mean - best_score - self.xi) / std)


__all__ = ["Acquisition"]



import math
import random

import numpy as np
import matplotlib.pyplot as plt

from abc import *
from typing import *

from .types import *


class DistributionBase(metaclass=ABCMeta):
    def __init__(
        self,
        lower: number_type = None,
        upper: number_type = None,
        *,
        values: List[Any] = None,
        **kwargs,
    ):
        number_types = (int, float)
        if lower is not None and not isinstance(lower, number_types):
            raise ValueError(f"lower should be a number, {type(lower)} found")
        if upper is not None and not isinstance(upper, number_types):
            raise ValueError(f"upper should be a number, {type(upper)} found")
        self.lower, self.upper, self.values, self.config = lower, upper, values, kwargs

    @property
    @abstractmethod
    def num_params(self) -> number_type:
        raise NotImplementedError

    @abstractmethod
    def pop(self) -> generic_number_type:
        raise NotImplementedError

    @abstractmethod
    def clip(self, value: generic_number_type) -> generic_number_type:
        pass

    @property
    def bounds(self) -> bounds_type:
        if self.values is None:
            return self.lower, self.upper
        return min(self.values), max(self.values)

    def __str__(self):
        if self.values is not None:
            return f"{type(self).__name__}[{', '.join(map(str, self.values))}]"
        if not self.config:
            cfg_str = ""
        else:
            sorted_key = sorted(self.config)
            cfg_str = f", {', '.join([f'{k}={self.config[k]}' for k in sorted_key])}"
        return f"{type(self).__name__}[{self.lower:.2f}, {self.upper:.2f}{cfg_str}]"

    __repr__ = __str__

    def _assert_lower_and_upper(self):
        assert self.lower is not None, "lower should be provided"
        assert self.upper is not None, "upper should be provided"

    def _assert_values(self):
        assert isinstance(self.values, list), "values should be a list"

    def visualize(self, n: int = 100) -> "DistributionBase":
        plt.figure()
        plt.scatter(list(range(n)), sorted(self.pop() for _ in range(n)))
        plt.show()
        return self


class Uniform(DistributionBase):
    @property
    def num_params(self) -> number_type:
        return math.inf

    def pop(self) -> number_type:
        self._assert_lower_and_upper()
        return random.random() * (self.upper - self.lower) + self.lower

    def clip(self, value: number_type) -> number_type:
        lower, upper = self.bounds
        return max(lower, min(upper, value))


class Exponential(Uniform):
    def __init__(
        self,
        lower: number_type = None,
        upper: number_type = None,
        *,
        values: List[Any] = None,
        **kwargs,
    ):
        super().__init__(lower, upper, values=values, **kwargs)
        self._assert_lower_and_upper()
        assert_msg = "lower should be greater than 0 in exponential distribution"
        assert self.lower > 0, assert_msg
        self.base = self.config.setdefault("base", 2)
        assert self.base > 1, "base should be greater than 1"
        self.lower, self.upper = map(
            math.log,
            [self.lower, self.upper],
            2 * [self.base],
        )

    @property
    def bounds(self) -> bounds_type:
        lower, upper = map(math.pow, 2 * [self.base], [self.lower, self.upper])
        return lower, upper

    def pop(self) -> number_type:
        return math.pow(self.base, super().pop())


class Choice(DistributionBase):
    @property
    def num_params(self) -> int:
        return len(self.values)

    def pop(self) -> Any:
        self._assert_values()
        return random.choice(self.values)

    def clip(self, value: generic_number_type) -> generic_number_type:
        if value is None or isinstance(value, str):
            return value
        diff = np.array([v - value for v in self.values], np.float32)
        best_idx = np.argmin(np.abs(diff)).item()
        return self.values[best_idx]


__all__ = ["DistributionBase", "Uniform", "Exponential", "Choice"]



import cv2
import math

import numpy as np

from typing import *
from collections import defaultdict
from scipy.optimize import linear_sum_assignment

from .utils import *
from ..misc import LoggingMixin, timing_context


class BBox(NamedTuple):
    # x, y 是中心点的坐标
    # angle 是以平面坐标系中 x 轴为 0°，y 轴为 90° 计的
    x: int
    y: int
    w: int
    h: int
    angle: float = 0.0

    @property
    def area(self) -> int:
        return self.w * self.h

    @property
    def center(self) -> Tuple[int, int]:
        return self.x, self.y

    @property
    def left_top(self) -> Tuple[int, int]:
        hw, hh = 0.5 * self.w, 0.5 * self.h
        ltx = -(hw * math.cos(self.angle) + hh * math.sin(self.angle))
        lty = hw * math.sin(self.angle) - hh * math.cos(self.angle)
        ltx, lty = map(int, [ltx + self.x, lty + self.y])
        return ltx, lty

    @property
    def right_bottom(self) -> Tuple[float, float]:
        hw, hh = 0.5 * self.w, 0.5 * self.h
        rbx = hw * math.cos(self.angle) + hh * math.sin(self.angle)
        rby = -(hw * math.sin(self.angle) - hh * math.cos(self.angle))
        rbx, rby = map(int, [rbx + self.x, rby + self.y])
        return rbx, rby

    def move(self, dx, dy) -> "BBox":
        new_x, new_y = self.x + dx, self.y + dy
        return BBox(new_x, new_y, self.w, self.h, self.angle)

    @classmethod
    def from_contour(cls, contour) -> "BBox":
        (x, y), (w, h), angle = cv2.minAreaRect(contour)
        return cls(*map(int, [x, y, w, h]), angle)

    @classmethod
    def from_contour_simplified(cls, contour) -> "BBox":
        x, y, w, h = cv2.boundingRect(contour)
        x, y, w, h = map(int, [x + 0.5 * w, y + 0.5 * h, w, h])
        return BBox(x, y, w, h, 0.0)


class ShapeObject(LoggingMixin):
    def __init__(
        self,
        gray: np.ndarray,
        contour: np.ndarray = None,
        *,
        force_hull: bool = False,
    ):
        self.gray = gray
        self._mask = None
        self._mask_area = None
        self._edge = None
        self._contour = contour
        self._force_hull = force_hull
        self._contour_center = None
        self._contour_area = None
        self._bbox = self._bbox_middle = self._bbox_simplified = None

    @property
    def wh(self) -> Tuple[int, int]:
        if len(self.gray.shape) == 2:
            h, w = self.gray.shape
        else:
            h, w, channels = self.gray.shape
        return w, h

    @property
    def canvas_area(self) -> float:
        w, h = self.wh
        return w * h

    @property
    def bbox(self) -> BBox:
        if self._bbox is None:
            self._bbox = BBox.from_contour(self.contour)
        return self._bbox

    @property
    def bbox_middle(self) -> BBox:
        if self._bbox_middle is None:
            center_y = self.contour_center[1]
            y_start, y_end = center_y - 2, center_y + 2
            middle_cropped = self.gray[y_start:y_end, ...]
            cropped_bbox = ShapeObject(middle_cropped).bbox_simplified
            self._bbox_middle = cropped_bbox.move(0, y_start)
        return self._bbox_middle

    @property
    def bbox_simplified(self) -> BBox:
        if self._bbox_simplified is None:
            self._bbox_simplified = BBox.from_contour_simplified(self.contour)
        return self._bbox_simplified

    @property
    def edge(self) -> np.ndarray:
        if self._edge is None:
            self._edge = cv2.Canny(self.gray, 50, 200)
        return self._edge

    @property
    def mask(self) -> np.ndarray:
        if self._mask is None:
            self._mask = self.gray >= 127
        return self._mask

    @property
    def mask_area(self) -> float:
        if self._mask_area is None:
            self._mask_area = self.mask.sum().item()
        return self._mask_area

    @property
    def contour(self) -> Union[np.ndarray, None]:
        if self._contour is None:
            self._contour = self.get_contour(self.gray, self._force_hull)
        return self._contour

    @property
    def contour_center(self) -> Tuple[float, float]:
        if self._contour_center is None:
            self._contour_center = self.get_contour_center(self.contour)
        return self._contour_center

    @property
    def contour_area(self) -> float:
        if self._contour_area is None:
            if self.contour is None or len(self.contour) == 0:
                self._contour_area = 0.0
            else:
                self._contour_area = self.get_contour_area(self.contour)
        return self._contour_area

    @property
    def is_valid(self) -> bool:
        return (
            self.contour is not None
            and self.contour_area > 0.0
            and self.mask_area < self.canvas_area
        )

    def copy(self) -> "ShapeObject":
        return ShapeObject(self.gray.copy(), self.contour.copy())

    def move(self, dx, dy) -> "ShapeObject":
        """
        平移移动图形，
        ------------> dx
        |
        |    坐标系
        |
        v dy

        :param dx: 在横轴的移动长度，dx为正的时候是右移，dx为负的时候为左移
        :param dy: 在纵轴的移动长度，dy为正的时候是下移，dy为负的时候为上移
        :return:
        """
        moved = self.move_img(self.gray, dx, dy)
        return ShapeObject(moved, (self.contour + [dx, dy]).astype(np.int))

    def to_tight(self, *, padding_ratio: float = 0.0) -> "ShapeObject":
        ltx, lty = self.bbox_simplified.left_top
        rbx, rby = self.bbox_simplified.right_bottom
        cropped = self.gray[lty:rby, ltx:rbx]
        cropped_contour = self.contour - [ltx, lty]
        if padding_ratio <= 0.0:
            return ShapeObject(cropped, cropped_contour)
        w, h = rbx - ltx, rby - lty
        x_padding = int(padding_ratio * w)
        y_padding = int(padding_ratio * h)
        new_w, new_h = w + 2 * x_padding, h + 2 * y_padding
        canvas = np.zeros([new_h, new_w], np.uint8)
        canvas[y_padding : y_padding + h, x_padding : x_padding + w] = cropped
        canvas_contour = cropped_contour + [x_padding, y_padding]
        return ShapeObject(canvas, canvas_contour)

    def to_center(self) -> "ShapeObject":
        w, h = self.wh
        current_center = self.contour_center
        target_center = w // 2, h // 2
        dx, dy = [tc - cc for tc, cc in zip(target_center, current_center)]
        return self.move(dx, dy)

    def match_center(
        self,
        template: "ShapeObject",
        *,
        use_bbox: bool = False,
        return_dxy: bool = False,
    ) -> Union["ShapeObject", Tuple[float, float]]:
        if not use_bbox:
            current_center = self.contour_center
            target_center = template.contour_center
        else:
            current_center = self.bbox_simplified.center
            target_center = template.bbox_simplified.center
        dx, dy = [tc - cc for tc, cc in zip(target_center, current_center)]
        if return_dxy:
            return dx, dy
        return self.move(dx, dy)

    def separate(
        self, *, min_area_threshold: float = 0.01, **kwargs
    ) -> List["ShapeObject"]:
        fg = self.gray >= 127
        bgx, bgy = np.vstack(np.nonzero(~fg))[..., 0]
        n_areas, areas_map = cv2.connectedComponents(self.gray)
        area_threshold = min_area_threshold * fg.sum()
        masks = [areas_map == i for i in range(n_areas) if i != areas_map[bgx, bgy]]
        thresholds = list(
            map(
                lambda mask: mask.astype(np.uint8) * 255,
                filter(lambda mask: mask.sum() > area_threshold, masks),
            )
        )
        return [ShapeObject(threshold, **kwargs) for threshold in thresholds]

    def get_scaled_img(self, scale: float, align_bottom: bool):
        return self.scale_img(
            self.gray,
            self.contour_center,
            scale,
            bbox=self.bbox_simplified,
            align_bottom=align_bottom,
        )

    def get_scaled_contour(
        self,
        w_scale,
        h_scale=None,
        *,
        canvas_changed=False,
    ) -> np.ndarray:
        if h_scale is None:
            h_scale = w_scale
        cx, cy = self.contour_center
        cnt_norm = self.contour - [cx, cy]
        cnt_scaled = cnt_norm * [w_scale, h_scale]
        if canvas_changed:
            cx, cy = cx * w_scale, cy * h_scale
        return (cnt_scaled + [cx, cy]).astype(np.int)

    @staticmethod
    def _fetch_match_shape_scores(src, tgt):
        scores = {}
        for method_num in [1, 2, 3]:
            score = cv2.matchShapes(src, tgt, method_num, 0.0)
            scores[str(method_num)] = max(0.0, 10.0 - score)
        return scores

    def get_match_scores(
        self,
        template: "ShapeObject",
        matching_method: Union[
            str, Callable, List[Union[str, Callable]]
        ] = "contour_all",
        **kwargs,
    ) -> Dict[str, float]:

        if kwargs.get("trigger_logging", False):
            self._init_logging()

        def _raise():
            raise NotImplementedError(
                f"matching_method '{matching_method}' not defined"
            )

        scores = defaultdict(float)
        matching_methods = matching_method
        if not isinstance(matching_method, (list, tuple)):
            matching_methods = [matching_method]

        for matching_method in matching_methods:
            with timing_context(self, matching_method):
                local_scores = {}
                if callable(matching_method):
                    local_scores["mean"] = matching_method(
                        self.contour, template.contour
                    )
                elif "contour" in matching_method:
                    if matching_method == "contour_all":
                        local_scores.update(
                            self._fetch_match_shape_scores(
                                self.contour, template.contour
                            )
                        )
                    elif matching_method == "contour_assignment":
                        with timing_context(self, "template resize"):
                            template_resized = template.resize(self.wh)
                        with timing_context(self, "get contour"):
                            self_contour, resized_contour = (
                                self.contour,
                                template_resized.contour,
                            )
                        with timing_context(self, "diff matrix"):
                            resized_transpose = resized_contour.transpose([1, 0, 2])
                            diff_matrix = self_contour - resized_transpose
                            diff_matrix = np.linalg.norm(diff_matrix, axis=-1)
                        with timing_context(self, "linear sum assignment"):
                            row_ind, col_ind = linear_sum_assignment(diff_matrix)
                        with timing_context(self, "diff from assignment"):
                            diff = diff_matrix[row_ind, col_ind].mean()
                            diff /= math.sqrt(self.wh[0] ** 2 + self.wh[1] ** 2)
                        local_scores["mean"] = 1.0 - diff
                    else:
                        _raise()
                elif "filled" in matching_method:
                    resized, template_resized = map(
                        ShapeObject.resize, [self, template]
                    )
                    gray, template_gray = resized.gray, template_resized.gray
                    if matching_method == "filled_shape_all":
                        local_scores.update(
                            self._fetch_match_shape_scores(gray, template_gray)
                        )
                    elif matching_method == "filled_iou":
                        local_scores["mean"] = 1.0 - self.iou_distance(
                            gray, template_gray
                        )
                    else:
                        _raise()
                elif matching_method == "edge":
                    # TODO : 通过缓存优化这里
                    num = kwargs.get("edge_scale_num", 1)
                    floor = kwargs.get("edge_scale_floor", 0.8)
                    ceiling = kwargs.get("edge_scale_ceiling", 0.8)
                    resize_floor = kwargs.get("edge_resize_floor", 0.5)
                    resized_template, resized_res = template.resize_to(
                        self,
                        False,
                        tight=True,
                        area_definition="bbox",
                    )
                    if resized_res["resize_scale"] <= resize_floor:
                        local_scores["mean"] = -math.inf
                    else:
                        for scale in np.linspace(floor, ceiling, num):
                            scaled_template = resized_template.rescale(
                                scale,
                                tight=True,
                                scale_contour=False,
                            )
                            match_score = cv2.matchTemplate(
                                self.edge,
                                scaled_template.edge,
                                cv2.TM_CCOEFF_NORMED,
                            ).max()
                            local_scores[f"{scale:6.4f}"] = match_score
                else:
                    _raise()
                for k, v in local_scores.items():
                    if k == "mean":
                        k = matching_method
                    scores[k] += v

        no_max, no_mean = "max" not in scores, "mean" not in scores
        if no_max or no_mean:
            max_score = max(scores.values())
            mean_score = sum(scores.values()) / len(scores)
            if no_max:
                scores["max"] = max_score
            if no_mean:
                scores["mean"] = mean_score
        return scores

    def get_min_distance(self, template: "ShapeObject") -> float:
        return np.min(
            np.linalg.norm(
                self.contour - template.contour[..., 0, :][None, ...],
                axis=-1,
            )
        )

    def resize(
        self,
        size: Tuple[float, float] = (300, 300),
        *,
        tight: bool = True,
        keep_aspect_ratio: bool = True,
    ) -> "ShapeObject":
        sw, sh = size
        if not tight:
            w, h = self.wh
            target = self
        else:
            tight = self.to_tight()
            w, h = tight.wh
            target = tight

        w_scale, h_scale = sw / w, sh / h
        if not keep_aspect_ratio:
            resized = cv2.resize(target.gray, size)
            resized_contour = target.get_scaled_contour(
                w_scale,
                h_scale,
                canvas_changed=True,
            )
        else:
            if w_scale < h_scale:
                adjust_x = False
                target_scale = w_scale
            else:
                adjust_x = True
                target_scale = h_scale
            resized_contour = target.get_scaled_contour(
                target_scale,
                canvas_changed=True,
            )
            target_size = tuple(
                map(int, map(round, [w * target_scale, h * target_scale]))
            )
            resized_raw = cv2.resize(target.gray, target_size)
            resized = np.zeros(size, dtype=np.uint8)
            tx, ty = target_size
            if adjust_x:
                dx = int(0.5 * (sw - tx))
                resized[:ty, dx : dx + tx] = resized_raw
                resized_contour += [dx, 0]
            else:
                dy = int(0.5 * (sh - ty))
                resized[dy : dy + ty, :tx] = resized_raw
                resized_contour += [0, dy]
        return ShapeObject(resized, resized_contour)

    def rescale(
        self,
        scale: float,
        *,
        tight: bool = False,
        align_bottom: bool = False,
        scale_contour: bool = True,
    ) -> "ShapeObject":
        dy, scaled = self.get_scaled_img(scale, align_bottom)
        contour = None if not scale_contour else self.get_scaled_contour(scale)
        if contour is not None and align_bottom:
            contour += [0, dy]
        scaled_obj = ShapeObject(scaled, contour)
        if tight:
            scaled_obj = scaled_obj.to_tight()
        return scaled_obj

    def resize_to(
        self,
        template: "ShapeObject",
        match_center: bool,
        *,
        tight: bool = False,
        scale_ceiling: int = None,
        area_definition: str = "contour",
    ) -> Union[None, Tuple["ShapeObject", Dict[str, float]]]:
        """
        将 self.contour 缩放，使得缩放后与 template.contour 的面积相同
        :param template：
        :param match_center：是否将返回的 ShapeObject 的中心点与 template 重合
        :param tight：是否将返回的 ShapeObject 的画布贴紧主体物
        :param scale_ceiling：放缩的最大比例，None 代指无限制
        :param area_definition：["contour", "bbox"]，面积的定义
        :return:
        """

        (w, h), (wt, ht) = self.wh, template.wh
        w_scale, h_scale = wt / w, ht / h
        resized = cv2.resize(self.gray, (wt, ht))
        new_obj = ShapeObject(
            resized, self.get_scaled_contour(w_scale, h_scale, canvas_changed=True)
        )
        if area_definition == "contour":
            template_area, new_area = template.contour_area, new_obj.contour_area
        elif area_definition == "bbox":
            template_area, new_area = template.bbox.area, new_obj.bbox.area
        else:
            raise NotImplementedError(
                f"area_definition '{area_definition}' is not defined"
            )
        scale = (template_area / new_area) ** 0.5
        if scale_ceiling is not None and scale >= scale_ceiling:
            return
        _, scaled = new_obj.get_scaled_img(scale, False)
        scaled_obj = ShapeObject(scaled, new_obj.get_scaled_contour(scale))
        if tight:
            if match_center:
                print("Warning : match_center will not take effect because tight=True")
            scaled_obj = scaled_obj.to_tight()
        elif match_center:
            scaled_obj = scaled_obj.match_center(template)
        return scaled_obj, {"resize_scale": scale}

    def rotate_to(
        self,
        template: "ShapeObject",
        rotated_per_itr: int = 10,
        distance_choice: str = "iou",
        match_center: bool = True,
        theta_floor: int = 0,
        theta_ceiling: int = 360,
        scale_ceiling: int = None,
    ) -> Union[None, Tuple["ShapeObject", Dict[str, float]]]:

        # 可以选择不同的distance function来定义匹配效果的好坏
        distance_function = {"iou": self.iou_distance}

        # 将 self.gray 缩放并将中心位置移到 template_img 的中心位置
        template_img = template.gray
        resize_result = self.resize_to(
            template, match_center, scale_ceiling=scale_ceiling
        )
        if resize_result is None:
            return
        resized, resize_res = resize_result
        center_pos = resized.contour_center

        if len(template_img.shape) == 2:
            rows, cols = template_img.shape
        else:
            rows, cols, channels = template_img.shape

        # 这里的iou_distance 是用1-iou_score后的，越小越好
        min_loss = distance_function[distance_choice](resized.gray, template_img)
        min_theta = 0
        rotated_img = resized.gray
        for theta in range(theta_floor, theta_ceiling, rotated_per_itr):
            m = cv2.getRotationMatrix2D(center_pos, theta, 1.0)
            # logger.info(type(M), M.shape)
            # 第三个参数是输出图像的尺寸中心
            tmp = cv2.warpAffine(resized.gray, m, (rows, cols))
            cur_loss = distance_function[distance_choice](template_img, tmp)
            if cur_loss < min_loss:
                min_loss = cur_loss
                min_theta = theta
                rotated_img = tmp
        res = {
            "theta": min_theta,
            "loss": min_loss,
            "resize_scale": resize_res["resize_scale"],
        }
        return ShapeObject(rotated_img), res

    def save_img(self, saving_path):
        cv2.imwrite(saving_path, self.gray)

    def save_edge(self, saving_path):
        cv2.imwrite(saving_path, self.edge)

    def save_contour(self, saving_path):
        w, h = self.wh
        draw_contours(w, h, [self.contour], saving_path=saving_path)

    @staticmethod
    def get_contour(gray, force_hull):
        """
        返回 gray 的轮廓，如果是单连通图像返回其外轮廓，如果是多个图形组成的返回其最小凸包
        :param gray: 512*512 的二值化图片
        :param force_hull: 是否强制使用最小凸包
        :return: 外形轮廓
        """
        contours, hierarchy = cv2.findContours(gray, mode=0, method=1)
        if not contours:
            return
        if len(contours) == 1 and not force_hull:
            hull = contours[0]
        else:
            all_cont = np.concatenate(contours, axis=0)
            hull = cv2.convexHull(all_cont)
        return hull

    @staticmethod
    def get_contour_area(contour):
        return cv2.contourArea(contour)

    @staticmethod
    def get_contour_center(contour):
        """
        获得形状轮廓的重心坐标
        :param contour:
        :return: 重心坐标
        """
        m = cv2.moments(contour)
        m00 = max(m["m00"], 1e-5)
        cx = int(m["m10"] / m00)
        cy = int(m["m01"] / m00)
        return cx, cy

    @staticmethod
    def scale_img(
        img,
        center,
        scale,
        *,
        bbox: BBox = None,
        align_bottom: bool = False,
    ):
        """
        对输入的图像，在其中心位置进行缩放
        :param img:
        :param center:
        :param scale:
        :param bbox: 若 align_bottom 为 True，则要提供一个 bbox
        :param align_bottom: 是否让 scale 后的 bbox 的 bottom 和原 bbox 的 bottom 对齐
        :return:
        """
        m = cv2.getRotationMatrix2D(center, angle=0, scale=scale)
        h, w = img.shape[:2]
        scaled = cv2.warpAffine(img, m, (w, h))
        if not align_bottom:
            return 0, scaled
        original_bottom_y = bbox.right_bottom[1]
        scaled_y = ShapeObject(scaled).bbox_simplified.right_bottom[1]
        dy = int(round(original_bottom_y - scaled_y))
        moved = ShapeObject.move_img(scaled, 0, dy)
        return dy, moved

    @staticmethod
    def move_img(img, dx, dy):
        h, w = img.shape[:2]
        m = np.array([[1, 0, dx], [0, 1, dy]], dtype=np.float)
        return cv2.warpAffine(img, m, (w, h))

    @staticmethod
    def iou_distance(gray1, gray2):
        m1, m2 = gray1 >= 127, gray2 >= 127
        return 1 - (m1 & m2).sum() / (m1 | m2).sum()

    @classmethod
    def from_file(cls, img_path, **kwargs):
        reverse = kwargs.pop("reverse", False)
        gray = Reader(img_path).to_gray(reverse=reverse)
        if gray is None:
            return
        threshold = cv2.threshold(gray, 127, 255, cv2.THRESH_BINARY)[1]
        return cls(threshold, **kwargs)


__all__ = ["BBox", "ShapeObject"]



cimport cython
cimport numpy as np
import numpy as np

from libc.math cimport isnan
from cython.parallel import prange


# api


@cython.wraparound(False)
@cython.boundscheck(False)
def rolling_sum(np.ndarray[np.float64_t, ndim=1] flat_data, int window, int mean):
    cdef int i, section_idx, cursor
    cdef unsigned int num_data = len(flat_data)
    cdef unsigned int num_sections = num_data // window

    cdef double nan = np.nan
    cdef double sum_cache, valid_sum_cache
    cdef np.ndarray[np.float64_t, ndim=1] results = np.empty(num_data, dtype=np.float64)

    for section_idx in prange(num_sections, nogil=True):
        cursor = window * section_idx

        sum_cache = valid_sum_cache = 0.0
        for i in range(window):
            if not isnan(flat_data[cursor + i]):
                sum_cache = sum_cache + flat_data[cursor + i]
                valid_sum_cache = valid_sum_cache + 1.0

        if valid_sum_cache == 0.0:
            results[cursor + window - 1] = nan
        else:
            if mean == 0:
                results[cursor + window - 1] = sum_cache
            else:
                results[cursor + window - 1] = sum_cache / valid_sum_cache
        for i in range(window - 1):
            if cursor + window + i < num_data:
                if not isnan(flat_data[cursor + window + i]):
                    sum_cache = sum_cache + flat_data[cursor + window + i]
                    valid_sum_cache = valid_sum_cache + 1.0
                if not isnan(flat_data[cursor + i]):
                    sum_cache = sum_cache - flat_data[cursor + i]
                    valid_sum_cache = valid_sum_cache - 1.0
                if valid_sum_cache == 0.0:
                    results[cursor + window + i] = nan
                else:
                    if mean == 0:
                        results[cursor + window + i] = sum_cache
                    else:
                        results[cursor + window + i] = sum_cache / valid_sum_cache
            else:
                break

    return results[window - 1:]


@cython.wraparound(False)
@cython.boundscheck(False)
def rolling_min(np.ndarray[np.float64_t, ndim=1] flat_data, int window):
    cdef int i, section_idx, cursor
    cdef unsigned int num_data = len(flat_data)
    cdef unsigned int num_sections = num_data // window

    cdef double cache_min, final_min
    cdef np.ndarray[np.float64_t, ndim=1] final_results = np.empty(num_data, dtype=np.float64)
    cdef np.ndarray[np.float64_t, ndim=1] caches = np.empty(num_data, dtype=np.float64)

    for section_idx in prange(num_sections, nogil=True):
        cursor = window * section_idx

        cache_min = flat_data[cursor + window - 1]
        for i in range(window - 1, -1, -1):
            if flat_data[cursor + i] < cache_min:
                cache_min = flat_data[cursor + i]
            caches[cursor + i] = cache_min

        final_min = flat_data[cursor + window]
        final_results[cursor + window - 1] = caches[cursor]
        for i in range(window - 1):
            if cursor + window + i < num_data:
                if flat_data[cursor + window + i] < final_min:
                    final_min = flat_data[cursor + window + i]
                final_results[cursor + window + i] = min(final_min, caches[cursor + i + 1])
            else:
                break

    return final_results[window - 1:]


@cython.wraparound(False)
@cython.boundscheck(False)
def rolling_max(np.ndarray[np.float64_t, ndim=1] flat_data, int window):
    cdef int i, section_idx, cursor
    cdef unsigned int num_data = len(flat_data)
    cdef unsigned int num_sections = num_data // window

    cdef double cache_max, final_max
    cdef np.ndarray[np.float64_t, ndim=1] final_results = np.empty(num_data, dtype=np.float64)
    cdef np.ndarray[np.float64_t, ndim=1] caches = np.empty(num_data, dtype=np.float64)

    for section_idx in prange(num_sections, nogil=True):
        cursor = window * section_idx

        cache_max = flat_data[cursor + window - 1]
        for i in range(window - 1, -1, -1):
            if flat_data[cursor + i] > cache_max:
                cache_max = flat_data[cursor + i]
            caches[cursor + i] = cache_max

        final_max = flat_data[cursor + window]
        final_results[cursor + window - 1] = caches[cursor]
        for i in range(window - 1):
            if cursor + window + i < num_data:
                if flat_data[cursor + window + i] > final_max:
                    final_max = flat_data[cursor + window + i]
                final_results[cursor + window + i] = max(final_max, caches[cursor + i + 1])
            else:
                break

    return final_results[window - 1:]


@cython.wraparound(False)
@cython.boundscheck(False)
def ema(np.ndarray[np.float64_t, ndim=1] flat_data, float ratio):
    cdef int i
    cdef unsigned int num_data = len(flat_data)
    cdef np.ndarray[np.float64_t, ndim=1] results = np.empty(num_data, dtype=np.float64)
    cdef double rev_ratio = 1.0 - ratio
    cdef double current, running

    for i in range(num_data):
        current = flat_data[i]
        if isnan(current):
            results[i] = current
        else:
            if i == 0:
                running = ratio * current
            else:
                running = ratio * current + rev_ratio * running
            results[i] = running

    return results



from abc import ABC, abstractmethod
from typing import Dict, List, Type, Union
from functools import partial

from ..types import *
from ..data_types import *
from ..distributions import *
from ....misc import register_core

union_data_type = Union[DataType, Iterable]
union_value_type = Union[generic_number_type, iterable_generic_number_type]
union_float_type = Union[float, List[float]]
union_bounds_type = Union[bounds_type, List[bounds_type]]
single_normalizer_dict: Dict[str, Type["SingleNormalizer"]] = {}


class SingleNormalizer(ABC):
    def __init__(self, data_type: DataType, **kwargs):
        self.core, self.dist = data_type, data_type.dist
        self._is_string = isinstance(data_type, String)
        self._is_exponential = isinstance(self.dist, Exponential)
        if not self.core.distribution_is_inf:
            self._choices = self.core.all()
        else:
            if not self.core.is_inf:
                self._choices = self.core.all()
            else:
                self._choices = None
                if not self._is_string:
                    self._lower, self._upper = self.core.lower, self.core.upper
                else:
                    self._lower, self._upper = self.dist.lower, self.dist.upper
                self._diff = self._upper - self._lower
        self._init_config(**kwargs)

    @property
    @abstractmethod
    def bounds(self) -> bounds_type:
        pass

    @abstractmethod
    def normalize(self, value: generic_number_type) -> float:
        pass

    @abstractmethod
    def recover(self, value: float) -> generic_number_type:
        pass

    def _init_config(self, **kwargs):
        pass

    @classmethod
    def register(cls, name: str):
        global single_normalizer_dict
        return register_core(name, single_normalizer_dict)


class IterableNormalizer:
    def __init__(
        self,
        data_type: Iterable,
        single_normalizer_base: Type[SingleNormalizer],
        **kwargs
    ):
        self._data_type = data_type
        self._normalizers = list(
            map(partial(single_normalizer_base, **kwargs), data_type.values)
        )

    @property
    def bounds(self) -> List[bounds_type]:
        return [normalizer.bounds for normalizer in self._normalizers]

    def normalize(self, value: iterable_generic_number_type) -> List[float]:
        return [
            normalizer.normalize(v) for normalizer, v in zip(self._normalizers, value)
        ]

    def recover(self, value: List[float]) -> iterable_generic_number_type:
        results = [
            normalizer.recover(v) for normalizer, v in zip(self._normalizers, value)
        ]
        if not self._data_type.is_list:
            results = tuple(results)
        return results


class Normalizer:
    def __init__(self, method: str, data_type: union_data_type, **kwargs):
        self._data_type = data_type
        self.is_iterable = isinstance(data_type, Iterable)
        single_normalizer_base = single_normalizer_dict[method]
        if not self.is_iterable:
            self._single_normalizer = single_normalizer_base(data_type, **kwargs)
        else:
            self._iterable_normalizer = IterableNormalizer(
                data_type,
                single_normalizer_base,
                **kwargs,
            )

    @property
    def bounds(self) -> union_bounds_type:
        if self.is_iterable:
            return self._iterable_normalizer.bounds
        return self._single_normalizer.bounds

    def normalize(self, value: union_value_type) -> union_float_type:
        if self.is_iterable:
            return self._iterable_normalizer.normalize(value)
        return self._single_normalizer.normalize(value)

    def recover(self, value: union_float_type) -> union_value_type:
        if self.is_iterable:
            return self._iterable_normalizer.recover(value)
        return self._single_normalizer.recover(value)


__all__ = ["Normalizer"]



import math
import numpy as np

from typing import Dict, List, Union, Iterator

from .types import *
from .data_types import *
from .normalizers import *
from .distributions import *
from ...misc import *

params_type = Dict[str, Union[DataType, Iterable, "params_type"]]


class ParamsGenerator:
    """
    Parameter generator for param searching, see cftool.ml.hpo.base.HPOBase for usage.

    Parameters
    ----------
    params : params_type, parameter settings.

    Examples
    ----------
    >>> grid = ParamsGenerator({
    >>>     "a": Any(Choice(values=[1, 2, 3])),
    >>>     "c": {
    >>>         "d": Int(Choice(values=[1, 2, 3])),
    >>>         "e": Float(Choice(values=[1, 2])),
    >>>     }
    >>> })
    >>> for param in grid.all():
    >>>     print(param)
    >>> # output : {'a': 1, 'c': {'d': 1, 'e': 1, 'f': 3}}, {'a': 1, 'c': {'d': 1, 'e': 1, 'f': 4}}
    >>> #          {'a': 1, 'c': {'d': 1, 'e': 2, 'f': 3}}, {'a': 1, 'c': {'d': 1, 'e': 2, 'f': 4}}
    >>> #          {'a': 1, 'c': {'d': 2, 'e': 1, 'f': 3}}, {'a': 1, 'c': {'d': 2, 'e': 1, 'f': 4}}
    >>> #          {'a': 1, 'c': {'d': 2, 'e': 2, 'f': 3}}, {'a': 1, 'c': {'d': 2, 'e': 2, 'f': 4}}
    >>> #          ......
    >>> #          {'a': 3, 'c': {'d': 3, 'e': 2, 'f': 3}}, {'a': 3, 'c': {'d': 3, 'e': 2, 'f': 4}}

    """

    def __init__(
        self,
        params: params_type,
        *,
        normalize_method: Union[str, None] = None,
        normalize_config: Dict[str, Any] = None,
    ):
        self._data_types = params

        def _data_type_offset(value: DataType) -> int:
            if not isinstance(value, Iterable):
                return 1
            return len(value.values)

        self._data_types_nested = Nested(params, offset_fn=_data_type_offset)

        if normalize_method is None:
            self._normalizers_flattened = None
        else:
            if normalize_config is None:
                normalize_config = {}

            def _data_type_normalizer(value: DataType) -> Normalizer:
                return Normalizer(normalize_method, value, **normalize_config)

            normalizers_nested = self._data_types_nested.apply(_data_type_normalizer)
            self._normalizers_flattened = normalizers_nested.flattened

        self._all_params_nested = self._all_flattened_data_types = None
        self._array_dim = self._all_bounds = None

    @property
    def params(self) -> params_type:
        return self._data_types

    @property
    def num_params(self) -> number_type:
        def _num_params(params):
            if isinstance(params, (DataType, Iterable)):
                return params.num_params
            assert isinstance(params, dict)
            num_params = prod(_num_params(v) for v in params.values())
            if math.isinf(num_params):
                return num_params
            return int(num_params)

        return _num_params(self._data_types)

    @property
    def array_dim(self) -> int:
        if self._array_dim is None:
            self._array_dim = self.flattened2array(
                self.flatten_nested(self.pop())
            ).shape[0]
        return self._array_dim

    @property
    def all_bounds(self) -> np.ndarray:
        if self._all_bounds is None:
            bounds_list = []
            for key in self.sorted_flattened_keys:
                if self._normalizers_flattened is None:
                    normalizer = None
                else:
                    normalizer = self._normalizers_flattened[key]
                if normalizer is None:
                    data_type = self._data_types_nested.get_value_from(key)
                    if not isinstance(data_type, Iterable):
                        bounds_list.append(list(data_type.bounds))
                    else:
                        bounds_list.extend(list(map(list, data_type.bounds)))
                else:
                    if normalizer.is_iterable:
                        bounds_list.extend(list(map(list, normalizer.bounds)))
                    else:
                        bounds_list.append(list(normalizer.bounds))
            self._all_bounds = np.array(bounds_list, np.float32)
        return self._all_bounds

    @property
    def all_flattened_params(self) -> all_flattened_type:
        if self._all_params_nested is None:
            apply = lambda data_type: data_type.all()
            self._all_params_nested = self._data_types_nested.apply(apply)
        return self._all_params_nested.flattened

    @property
    def sorted_flattened_keys(self) -> List[str]:
        return self._data_types_nested.sorted_flattened_keys

    def pop(self) -> nested_type:
        def _pop(src: dict, tgt: dict):
            for k, v in src.items():
                if isinstance(v, dict):
                    next_tgt = tgt.setdefault(k, {})
                    _pop(v, next_tgt)
                else:
                    tgt[k] = v.pop()
            return tgt

        return _pop(self._data_types, {})

    def all(self) -> Iterator[nested_type]:
        for flattened_params in Grid(self.all_flattened_params):
            yield self._data_types_nested.nest_flattened(flattened_params)

    def flatten_nested(self, nested: nested_type) -> nested_type:
        return self._data_types_nested.flatten_nested(nested)

    def nest_flattened(self, flattened: flattened_type) -> nested_type:
        return self._data_types_nested.nest_flattened(flattened)

    def flattened2array(self, flattened: flattened_type) -> np.ndarray:
        if self._normalizers_flattened is None:
            normalized = flattened
        else:
            normalized = {
                k: self._normalizers_flattened[k].normalize(v)
                for k, v in flattened.items()
            }
        return self._data_types_nested.flattened2array(normalized)

    def array2flattened(self, array: np.ndarray) -> flattened_type:
        normalized = self._data_types_nested.array2flattened(array)
        if self._normalizers_flattened is None:
            flattened = normalized
        else:
            flattened = {
                k: self._normalizers_flattened[k].recover(v)
                for k, v in normalized.items()
            }
        for key, value in flattened.items():
            data_type = self._data_types_nested.get_value_from(key)
            flattened[key] = data_type.transform(value)
        return flattened


__all__ = ["ParamsGenerator", "params_type"]



import warnings

from typing import *
from tqdm import tqdm
from sklearn.gaussian_process.kernels import Matern
from sklearn.gaussian_process import GaussianProcessRegressor

from .acquisition import *
from .target_space import *
from ...misc import *
from ...ml.param_utils import params_type


class BayesianOptimization:
    def __init__(
        self,
        fn: fn_type,
        params: params_type,
        *,
        gp_params: Dict[str, Any] = None,
        acquisition: str = "ucb",
        normalization: Union[str, None] = "cube",
        normalization_config: Dict[str, Any] = None,
        xi: float = 0.01,
        kappa: float = 2.0,
        kappa_decay: float = 1.0,
        kappa_decay_delay: int = 0,
    ):

        self._queue = []
        self._space = TargetSpace(
            fn,
            params,
            normalization=normalization,
            normalization_config=normalization_config,
        )

        if gp_params is None:
            gp_params = {}
        kernel_params = gp_params.pop("kernel_params", {})
        kernel_params.setdefault("nu", 2.5)
        gp_params.setdefault("alpha", 1e-6)
        gp_params.setdefault("normalize_y", True)
        gp_params.setdefault("n_restarts_optimizer", 5)
        kernel = Matern(**kernel_params)
        self._gp = GaussianProcessRegressor(kernel=kernel, **gp_params)

        self._acquisition = Acquisition(
            self._gp,
            acquisition,
            xi,
            kappa,
            kappa_decay,
            kappa_decay_delay,
        )

    @property
    def space(self) -> TargetSpace:
        return self._space

    @property
    def best_result(self) -> Result:
        return self._space.best_result

    @property
    def all_results(self) -> List[Result]:
        return self._space.all_results

    def register(self, param: flattened_type, score: float) -> "BayesianOptimization":
        self._space.register(param, score)
        return self

    def probe(
        self,
        param: flattened_type,
        *,
        to_queue: bool = True,
    ) -> "BayesianOptimization":
        if to_queue:
            self._queue.append(param)
        else:
            self._space.probe(param)
        return self

    def suggest(self, num_warmup: int, num_iter: int) -> flattened_type:
        if self._space.is_empty:
            return self._space.sample()

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self._gp.fit(self._space.tried_flattened_params, self._space.tried_scores)

        suggestion = self._acquisition.search_max(
            self._space.tried_scores.max(),
            self._space,
            num_warmup,
            num_iter,
        )

        return self._space.array2param(suggestion)

    def maximize(
        self,
        *,
        init_points: int = 5,
        num_epoch: int = 20,
        num_warmup: int = 10000,
        num_iter: int = 10,
        use_tqdm: bool = True,
    ) -> "BayesianOptimization":

        if not self._queue and self._space.is_empty:
            init_points = max(init_points, 1)
        for _ in range(init_points):
            self._queue.append(self._space.sample())

        counter = 0
        if not use_tqdm:
            iterator = None
        else:
            iterator = tqdm(list(range(len(self._queue) + num_epoch)))
        while self._queue or counter < num_epoch:
            if self._queue:
                x_probe = self._queue.pop(0)
            else:
                self._acquisition.update()
                x_probe = self.suggest(num_warmup, num_iter)
                counter += 1
            self.probe(x_probe, to_queue=False)
            if iterator is not None:
                iterator.update()

        return self


__all__ = ["BayesianOptimization"]



from typing import *

from ..base import HPOBase
from ...utils import Metrics
from ....optim import BayesianOptimization


@HPOBase.register("bo")
class BayesianHPO(HPOBase):
    @property
    def is_sequential(self) -> bool:
        return True

    def _init_config(self, **kwargs):
        self._bo_config = kwargs.get("bo_config", {})
        bo_normalization = self._bo_config.setdefault("normalization", "cube")
        if bo_normalization == "cube":
            bo_norm_cfg = self._bo_config.setdefault("normalization_config", {})
            bo_norm_cfg.setdefault("convert_only", False)
        self._num_iter = kwargs.get("num_iter", 10)
        self._num_warmup = kwargs.get("num_warmup", 10000)
        self._init_points = kwargs.get("init_points", 5)
        if self._init_points <= 1:
            msg = f"init_points should larger than 1, {self._init_points} found"
            raise ValueError(msg)
        self._bo_core, self._iteration = None, 0

    def _score(self, final_scores: Dict[str, float]) -> float:
        return sum(
            [
                self._score_weights.setdefault(k, 1.0) * v * Metrics.sign_dict[k]
                for k, v in final_scores.items()
            ]
        )

    def _sample_param(self) -> Union[None, Dict[str, Any]]:
        self._iteration += 1
        if self._bo_core is None:
            params = self.param_generator.params
            self._bo_core = BayesianOptimization(None, params, **self._bo_config)
        if self._iteration <= self._init_points:
            return self.param_generator.pop()
        if not self._bo_core.space.is_empty:
            nested = self.last_param
            flattened = self.param_generator.flatten_nested(nested)
            self._bo_core.register(
                flattened,
                self._score(self._get_scores(self.last_patterns)),
            )
        else:
            for code, params in self.param_mapping.items():
                patterns = self.patterns[code]
                flattened = self.param_generator.flatten_nested(params)
                self._bo_core.register(
                    flattened,
                    self._score(self._get_scores(patterns)),
                )
        flattened = self._bo_core.suggest(self._num_warmup, self._num_iter)
        return self.param_generator.nest_flattened(flattened)


__all__ = ["BayesianHPO"]



import math
import pprint
import random
import logging

import numpy as np

from typing import *
from tqdm import tqdm
from functools import partial
from collections import defaultdict
from abc import abstractmethod, ABCMeta

from ..utils import *
from ...misc import *
from ..param_utils import *
from ...dist import Parallel

hpo_dict: Dict[str, Type["HPOBase"]] = {}
created_type = Union[pattern_type, Any]
creator_type = Callable[[np.ndarray, np.ndarray, Dict[str, Any]], created_type]


class HPOBase(LoggingMixin, metaclass=ABCMeta):
    def __init__(
        self,
        creator: creator_type,
        params: Dict[str, DataType],
        *,
        converter: Callable[[List[created_type]], patterns_type] = None,
        verbose_level: int = 2,
        **kwargs,
    ):
        self._caches = {}
        self._init_config(**kwargs)
        self._creator = creator
        self._converter = converter
        self.param_generator = ParamsGenerator(params)
        self._verbose_level = verbose_level

    @property
    @abstractmethod
    def is_sequential(self) -> bool:
        pass

    @property
    def last_param(self) -> nested_type:
        return self.param_mapping[self.last_code]

    @property
    def last_patterns(self) -> List[pattern_type]:
        return self.patterns[self.last_code]

    def _init_config(self, **kwargs):
        pass

    def _sample_param(self) -> Union[None, nested_type]:
        if self.is_sequential:
            raise NotImplementedError
        return

    def _get_scores(self, patterns: List[pattern_type]) -> Dict[str, float]:
        key = "core"
        comparer = Comparer({key: patterns}, self.estimators)
        final_scores = comparer.compare(
            self.x_validation,
            self.y_validation,
            scoring_function=self._estimator_scoring_function,
            verbose_level=6,
        ).final_scores
        return {metric: scores[key] for metric, scores in final_scores.items()}

    def _core(
        self,
        param: nested_type,
        *,
        convert: bool = True,
        parallel_run: bool = False,
    ) -> List[created_type]:
        range_list = list(range(self._num_retry))
        _task = lambda _=0: self._creator(self.x_train, self.y_train, param)
        tqdm_config = {"position": 1, "leave": False}
        if not parallel_run:
            if self._use_tqdm and len(range_list) > 1:
                range_list = tqdm(range_list, **tqdm_config)
            created = [_task() for _ in range_list]
        else:
            parallel = Parallel(
                self._num_jobs,
                use_tqdm=self._use_tqdm,
                tqdm_config=tqdm_config,
            )
            created = parallel(_task, range_list).ordered_results
        if not convert:
            return created
        patterns = created if self._converter is None else self._converter(created)
        return patterns

    def search(
        self,
        x: generic_data_type,
        y: generic_data_type,
        estimators: List[Estimator],
        x_validation: generic_data_type = None,
        y_validation: generic_data_type = None,
        *,
        num_jobs: int = 4,
        num_retry: int = 4,
        num_search: Union[str, int, float] = 10,
        score_weights: Union[Dict[str, float], None] = None,
        estimator_scoring_function: Union[str, scoring_fn_type] = "default",
        use_tqdm: bool = True,
        verbose_level: int = 2,
        **kwargs,
    ) -> "HPOBase":

        if x_validation is None and y_validation is None:
            x_validation, y_validation = x, y

        self.estimators = estimators
        self.x_train, self.y_train = x, y
        self.x_validation, self.y_validation = x_validation, y_validation

        num_params = self.param_generator.num_params
        if isinstance(num_search, str):
            if num_search != "all":
                raise ValueError(
                    "num_search can only be 'all' when it is a string, "
                    f"'{num_search}' found"
                )
            if num_params == math.inf:
                raise ValueError(
                    "num_search is 'all' but we have infinite params to search"
                )
            num_search = num_params
        if num_search > num_params:
            self.log_msg(
                f"`n` is larger than total choices we've got ({num_params}), "
                f"therefore only {num_params} searches will be run",
                self.warning_prefix,
                msg_level=logging.WARNING,
            )
            num_search = num_params
        num_jobs = min(num_search, num_jobs)

        self._use_tqdm = use_tqdm
        if score_weights is None:
            score_weights = {estimator.type: 1.0 for estimator in estimators}
        self._score_weights = score_weights
        self._estimator_scoring_function = estimator_scoring_function
        self._num_retry, self._num_jobs = num_retry, num_jobs

        with timeit("Generating Patterns"):
            if self.is_sequential:
                self.patterns, self.param_mapping = {}, {}
                iterator = list(range(num_search))
                if use_tqdm:
                    iterator = tqdm(iterator, position=0)
                for _ in iterator:
                    param = self._sample_param()
                    self.last_code = hash_code(str(param))
                    self.param_mapping[self.last_code] = param
                    self.patterns[self.last_code] = self._core(
                        param, convert=True, parallel_run=num_jobs > 1
                    )
            else:
                if num_params == math.inf:
                    all_params = [self.param_generator.pop() for _ in range(num_search)]
                else:
                    all_params = []
                    sampled = random.sample(list(range(num_search)), k=num_search)
                    all_indices = set(sampled)
                    for i, param in enumerate(self.param_generator.all()):
                        if i in all_indices:
                            all_params.append(param)
                        if len(all_params) == num_search:
                            break

                codes = list(map(hash_code, map(str, all_params)))
                self.param_mapping = dict(zip(codes, all_params))
                if num_jobs <= 1:
                    if self._use_tqdm:
                        all_params = tqdm(all_params)
                    patterns = list(map(self._core, all_params))
                else:
                    logging_folder = kwargs.get("parallel_logging_folder")
                    parallel = Parallel(num_jobs, logging_folder=logging_folder)
                    core = partial(self._core, convert=False)
                    created_list = parallel(core, all_params).ordered_results
                    patterns = list(map(self._converter, created_list))
                self.patterns = dict(zip(codes, patterns))
                self.last_code = codes[-1]

        self.comparer = Comparer(self.patterns, estimators)
        self.comparer.compare(
            x_validation,
            y_validation,
            scoring_function=estimator_scoring_function,
            verbose_level=verbose_level,
        )

        weighted_scores = defaultdict(float)
        for metric, scores in self.comparer.final_scores.items():
            for method, score in scores.items():
                weighted_scores[method] += self._score_weights[metric] * score
        sorted_methods = sorted(weighted_scores)
        sorted_methods_scores = [weighted_scores[key] for key in sorted_methods]
        best_method = sorted_methods[np.argmax(sorted_methods_scores).item()]
        self.best_param = self.param_mapping[best_method]

        best_methods = self.comparer.best_methods
        self.best_params = {k: self.param_mapping[v] for k, v in best_methods.items()}
        param_msgs = {k: pprint.pformat(v) for k, v in self.best_params.items()}
        estimator_statistics = self.comparer.estimator_statistics
        sorted_metrics = sorted(param_msgs)
        target_statistics = []
        for metric in sorted_metrics:
            metric_method = best_methods[metric]
            metric_statistics = estimator_statistics[metric][metric_method]
            target_statistics.append(
                {
                    "method": metric_method,
                    "mean": fix_float_to_length(metric_statistics["mean"], 8),
                    "std": fix_float_to_length(metric_statistics["std"], 8),
                }
            )
        msg = "\n".join(
            sum(
                [
                    [
                        "-" * 100,
                        f"{metric}  ({stat['method']}) "
                        f"({stat['mean']} ± {stat['std']})",
                        "-" * 100,
                        param_msgs[metric],
                    ]
                    for metric, stat in zip(sorted_metrics, target_statistics)
                ],
                [],
            )
            + [
                "-" * 100,
                f"best ({best_method})",
                "-" * 100,
                pprint.pformat(self.best_param),
            ]
            + ["-" * 100]
        )
        self.log_block_msg(msg, self.info_prefix, "Best Parameters", verbose_level)

        return self

    @staticmethod
    def make(method: str, *args, **kwargs) -> "HPOBase":
        return hpo_dict[method](*args, **kwargs)

    @classmethod
    def register(cls, name):
        global hpo_dict
        return register_core(name, hpo_dict)


__all__ = ["HPOBase", "hpo_dict"]



import numpy as np
import matplotlib.pyplot as plt

from tqdm import tqdm
from typing import *
from functools import partial
from abc import ABC, abstractmethod

from ...misc import register_core, fix_float_to_length

optimizer_dict: Dict[str, Type["Optimizer"]] = {}


class Optimizer(ABC):
    def __init__(self, lr, **kwargs):
        self._lr = lr
        self._caches = {}

    @abstractmethod
    def step(
        self,
        model: "GradientDescentMixin",
        gradient_dict: Dict[str, np.ndarray],
    ):
        pass

    def reset(self):
        self._caches.clear()

    def apply(self, key, gradient, model):
        attr = getattr(model, key)
        attr -= self._lr * gradient
        setattr(model, key, attr)

    @classmethod
    def register(cls, name):
        global optimizer_dict
        return register_core(name, optimizer_dict)


class GradientDescentMixin(ABC):
    @property
    def lr(self):
        return getattr(self, "_lr", 0.01)

    @property
    def opt(self):
        return getattr(self, "_opt", "sgd")

    @property
    def epoch(self):
        return getattr(self, "_epoch", 20)

    @property
    def batch_size(self):
        return getattr(self, "_batch_size", 32)

    @property
    def optimizer_config(self):
        return getattr(self, "_optimizer_config", {})

    @property
    def show_tqdm(self):
        return getattr(self, "_show_tqdm", True)

    @show_tqdm.setter
    def show_tqdm(self, value: bool):
        self._show_tqdm = value

    @abstractmethod
    def parameter_names(self) -> List[str]:
        """
        This method returns all parameters' names, each name should
        correspond to a property.

        e.g. in LinearRegression, if self._w & self._b are the parameters,
        this method should return : ["_w", "_b"].

        Returns
        -------
        names : List[str]

        """

    @abstractmethod
    def loss_function(
        self,
        x_batch: np.ndarray,
        y_batch: np.ndarray,
        batch_indices: np.ndarray,
    ) -> Dict[str, Any]:
        """
        This method calculate the loss of one batch.

        Parameters
        ----------
        x_batch : np.ndarray, one batch of training set features.
        y_batch : np.ndarray, one batch of training set labels.
        batch_indices : np.ndarray, indices of current batch.

        Returns
        -------
        results : Dict[str, Any], contains loss value and intermediate results
                                  which are critical.
        * Must contains 'loss' key, whose value should be a float.

        """

    @abstractmethod
    def gradient_function(
        self,
        x_batch: np.ndarray,
        y_batch: np.ndarray,
        batch_indices: np.ndarray,
        loss_dict: Dict[str, Any],
    ) -> Dict[str, np.ndarray]:
        """
        This method calculate the gradients of one batch.

        Parameters
        ----------
        x_batch : np.ndarray, one batch of training set features.
        y_batch : np.ndarray, one batch of training set labels.
        batch_indices : np.ndarray, indices of current batch.
        loss_dict : Dict[str, Any], results from self.loss_function.

        Returns
        -------
        results : Dict[str, np.ndarray], contains all gradients needed.
        * set(results.keys()) should be identical with set(self.parameter_names)

        """

    def _setup_optimizer(self, **kwargs):
        if getattr(self, "_optimizer", None) is None:
            self._optimizer = optimizer_dict[self.opt](self.lr, **kwargs)

    def setup_optimizer(
        self,
        optimizer: str,
        lr: float,
        *,
        epoch: int = 20,
        batch_size: int = 32,
        **kwargs
    ) -> "GradientDescentMixin":
        self._lr = lr
        self._opt = optimizer
        self._epoch = epoch
        self._batch_size = batch_size
        self._setup_optimizer(**kwargs)
        return self

    def gradient_descent(self, x: np.ndarray, y: np.ndarray) -> "GradientDescentMixin":
        self._setup_optimizer(**self.optimizer_config)
        self._optimizer.reset()
        n_sample = len(x)
        b_size = min(n_sample, self.batch_size)
        n_step = n_sample // b_size
        n_step += int(n_step * b_size < n_sample)
        iterator = range(self.epoch)
        if self.show_tqdm:
            iterator = tqdm(iterator, total=self.epoch)
        self._losses = []
        for _ in iterator:
            local_losses = []
            indices = np.random.permutation(n_sample)
            for j in range(n_step):
                batch_indices = indices[j * b_size : (j + 1) * b_size]
                x_batch, y_batch = x[batch_indices], y[batch_indices]
                loss_dict = self.loss_function(x_batch, y_batch, batch_indices)
                gradient_dict = self.gradient_function(
                    x_batch, y_batch, batch_indices, loss_dict
                )
                self._optimizer.step(self, gradient_dict)
                local_losses.append(loss_dict["loss"])
            if self.show_tqdm:
                iterator.set_postfix({"loss": fix_float_to_length(local_losses[0], 6)})
            self._losses.append(local_losses)
        return self

    def plot_loss_curve(self) -> "GradientDescentMixin":
        base = np.arange(len(self._losses))
        to_array = partial(np.array, dtype=np.float32)
        losses = list(map(to_array, self._losses))
        mean, std = map(
            to_array,
            map(list, [map(np.mean, losses), map(np.std, losses)]),
        )
        ax = plt.gca()
        ax.plot(base, mean)
        plt.title("loss curve")
        ax.fill_between(base, mean + std, mean - std, color="#b9cfe7")
        plt.show()
        return self


__all__ = ["Optimizer", "GradientDescentMixin", "optimizer_dict"]



import psutil
import logging
import subprocess

from typing import *

from .misc import PureLoggingMixin


class PCManager:
    """ Util class which can check PC status (unit: MB). """

    @staticmethod
    def get_available_ram() -> float:
        return psutil.virtual_memory().available / 1024 ** 2

    @staticmethod
    def get_pid_ram_usage(pid: int) -> float:
        try:
            process = psutil.Process(pid)
        except psutil.NoSuchProcess:
            return 0.0
        return process.memory_info().rss / 1024 ** 2


class GPUManager:
    """
    Util class which can check CUDA usages.

    Parameters
    ----------
    available_cuda_list : {None, list}, indicates CUDAs which are available.
    * If None, then all CUDAs will be available.
    reuse : bool, indicates whether one CUDA could be used multiple times.
    * If `available_cuda_list` is None, then `reuse` will have no effect.

    Examples
    --------
    >>> available_cuda_list = list(range(8))  # indicates that CUDA-0, ..., CUDA-7 is available
    >>> cuda = GPUManager(available_cuda_list).choose()  # `cuda` will be the CUDA id with most memory

    """

    def __init__(
        self,
        available_cuda_list: Optional[List[Union[str, int]]] = None,
        *,
        reuse: bool = True,
    ):
        self._reuse = reuse
        if available_cuda_list is None:
            self._available_cuda = None
        else:
            self._available_cuda = set(map(int, available_cuda_list))

    @staticmethod
    def _query_gpu() -> List[Dict[str, int]]:
        q_args = ["index", "gpu_name", "memory.free", "memory.total"]
        cmd = f"nvidia-smi --query-gpu={','.join(q_args)} --format=csv,noheader"
        results = subprocess.Popen(
            cmd,
            shell=True,
            stdout=subprocess.PIPE,
            close_fds=True,
        ).communicate()[0]
        results = list(filter(bool, results.decode("utf-8").strip().split("\n")))

        def _parse(line):
            numeric_args = ["memory.free", "memory.total"]
            power_manage_enable = lambda v: ("Not Support" not in v)
            to_numeric = lambda v: float(
                v.upper().strip().replace("MIB", "").replace("W", "")
            )
            process = lambda k, v: (
                (int(to_numeric(v)) if power_manage_enable(v) else 1)
                if k in numeric_args
                else v.strip()
            )
            rs = {k: process(k, v) for k, v in zip(q_args, line.strip().split(","))}
            rs["index"] = int(rs["index"])
            return rs

        return list(map(_parse, results))

    @staticmethod
    def _query_pid() -> List[Tuple[int, int]]:
        cmd = "nvidia-smi --query-compute-apps=pid,used_memory --format=csv,noheader"
        results = subprocess.Popen(
            cmd,
            shell=True,
            stdout=subprocess.PIPE,
            close_fds=True,
        ).communicate()[0]
        results = list(filter(bool, results.decode("utf-8").strip().split("\n")))
        pid_list = []
        for line in results:
            key, value = map(int, line.replace("MiB", "").strip().split(", "))
            pid_list.append((key, value))
        return pid_list

    @staticmethod
    def _sort_by_memory(
        gpus: List[Dict[str, int]],
        by_size: bool = False,
    ) -> List[Dict[str, int]]:
        if by_size:
            return sorted(gpus, key=lambda d: d["memory.free"], reverse=True)
        return sorted(
            gpus,
            key=lambda d: float(d["memory.free"]) / d["memory.total"],
            reverse=True,
        )

    def _gpu_filter(self, gpu: Dict[str, int]):
        return self._available_cuda is None or gpu["index"] in self._available_cuda

    def choose(self) -> int:
        if isinstance(self._available_cuda, set) and not self._available_cuda:
            raise ValueError("No more CUDAs are available")
        chosen_gpu = int(
            next(
                filter(self._gpu_filter, self._sort_by_memory(self._query_gpu(), True))
            )["index"]
        )
        if not self._reuse and self._available_cuda is not None:
            self._available_cuda.remove(chosen_gpu)
        return chosen_gpu

    def get_gpu_frees(self) -> Dict[int, int]:
        return {
            gpu["index"]: gpu["memory.free"]
            for gpu in filter(self._gpu_filter, self._query_gpu())
        }

    def get_pid_usages(self) -> Dict[int, int]:
        return dict(self._query_pid())


class ResourceManager:
    """
    Util class which can monitor & manage resources.
    * It utilizes `PCManager` & `GPUManager` defined above.
    * It is currently used in cftool.dist.core.Parallel only.

    """

    def __init__(
        self,
        config: Dict[str, Any],
        get_task_name: Callable[[int], Union[str, None]],
        refresh_patience: int,
    ):
        self._resources, self._info_dict, self._overwritten_task_info = [], {}, {}
        self._init_logger = self._meta_log_name = None
        self._log_msg = self._log_block_msg = None
        self._log_meta_msg = self._log_with_meta = None
        self.pid2task_idx, self._get_task_name = {}, get_task_name
        self.config, self._refresh_patience = config, refresh_patience

    def register(self, resource_name: str, methods: Dict[str, Callable]) -> None:
        self._resources.append(resource_name)
        resource_config = self.config.setdefault(f"{resource_name.lower()}_config", {})
        preset_usages = resource_config.setdefault("preset_usages", {})
        minimum_resource = resource_config.setdefault(
            "minimum_resource",
            preset_usages.setdefault("__default__", 1024),
        )
        counter_threshold = resource_config.setdefault("counter_threshold", 4)
        warning_threshold = resource_config.setdefault("warning_threshold", 1024)
        info = self._info_dict.setdefault(
            resource_name,
            {
                "preset_usages": preset_usages,
                "minimum_resource": minimum_resource,
                "counter_threshold": counter_threshold,
                "warning_threshold": warning_threshold,
            },
        )
        get_pid_usage, get_pid_usage_dict, get_available_dict = map(
            methods.get,
            ["get_pid_usage", "get_pid_usage_dict", "get_available_dict"],
        )
        if get_pid_usage is None and get_pid_usage_dict is None:
            raise ValueError(
                "either get_pid_usage or get_pid_usage_dict "
                "should be provided in methods"
            )
        if get_available_dict is None:
            raise ValueError("get_available_dict should be provided in methods")
        info["get_pid_usage"] = get_pid_usage
        info["get_pid_usage_dict"] = get_pid_usage_dict
        info["get_available_dict"] = get_available_dict
        info["inference_frees"], info["checkpoint_pid_usages"] = {}, {}
        info["running_pid_usages"], info["running_pid_counters"] = {}, {}
        info["inference_usages_initialized"] = False
        info["pid2resource_id"] = {}

    def register_logging(
        self,
        init_logger: Callable[[str], None],
        mixin: PureLoggingMixin,
    ):
        self._init_logger, self._meta_log_name = init_logger, mixin.meta_log_name
        self._log_msg, self._log_block_msg = mixin.log_msg, mixin.log_block_msg
        self._log_meta_msg = mixin._log_meta_msg
        self._log_with_meta = mixin._log_with_meta

    @property
    def inference_usages_initialized(self) -> bool:
        for info in self._info_dict.values():
            if not info["inference_usages_initialized"]:
                return False
        return True

    @property
    def checkpoint_initialized(self) -> bool:
        for info in self._info_dict.values():
            if not info["checkpoint_pid_usages"]:
                return False
        return True

    @staticmethod
    def _get_all_relevant_processes(pid: int) -> List[psutil.Process]:
        parent = psutil.Process(pid)
        processes = [parent]
        for process in parent.children(recursive=True):
            processes.append(process)
        return processes

    @staticmethod
    def _get_pid_usages(info: Dict[str, Any]) -> Dict[str, int]:
        get_pid_usage_dict = info["get_pid_usage_dict"]
        if get_pid_usage_dict is not None:
            return get_pid_usage_dict()
        pid_usages, get_pid_usage = {}, info["get_pid_usage"]
        for pid in info["running_pid_usages"].keys():
            try:
                processes = ResourceManager._get_all_relevant_processes(pid)
                pid_usages[pid] = int(
                    sum(get_pid_usage(process.pid) for process in processes)
                )
            except psutil.NoSuchProcess:
                pid_usages[pid] = 0
        return pid_usages

    @staticmethod
    def get_dict_block_msg(d: Dict[str, int]) -> str:
        keys = sorted(d.keys())
        values = [d[key] for key in keys]
        header = " | ".join(map(lambda pid: f"{pid:^12s}", map(str, keys)))
        body = " | ".join(map(lambda usage: f"{int(usage):^10d}MB", values))
        len_header = len(header)
        above = f"{'=' * len_header}\n{header}\n{'-' * len_header}"
        return f"{above}\n{body}\n{'-' * len_header}"

    def default_usage(self, resource: str) -> int:
        return int(self._info_dict[resource]["preset_usages"]["__default__"])

    def initialize_running_usages(self) -> None:
        for info in self._info_dict.values():
            info["running_pid_usages"] = info["checkpoint_pid_usages"].copy()

    def initialize_inference_usages(self) -> None:
        for info in self._info_dict.values():
            info["inference_frees"] = info["get_available_dict"]()
            info["inference_usages_initialized"] = True

    def log_pid_usages_and_inference_frees(self) -> None:
        for resource in self._resources:
            info = self._info_dict[resource]
            self._log_block_msg(
                self._meta_log_name,
                f"current pid {resource} usages",
                self.get_dict_block_msg(self._get_pid_usages(info)),
                logging.DEBUG,
            )
            self._log_block_msg(
                self._meta_log_name,
                f"current inference {resource} frees",
                self.get_dict_block_msg(info["inference_frees"]),
                logging.DEBUG,
            )

    def check(self) -> None:
        for resource in self._resources:
            info = self._info_dict[resource]
            checkpoint_pid_usages, running_pid_usages, running_pid_counters = map(
                info.get,
                [
                    "checkpoint_pid_usages",
                    "running_pid_usages",
                    "running_pid_counters",
                ],
            )
            get_pid_usage, get_pid_usage_dict = map(
                info.get,
                ["get_pid_usage", "get_pid_usage_dict"],
            )
            if get_pid_usage is None:
                pid_usage_dict = get_pid_usage_dict()
                get_pid_usage = lambda pid_: pid_usage_dict.get(pid_, 0)
            for pid, checkpoint_usage in checkpoint_pid_usages.items():
                task_name = self._get_task_name(self.pid2task_idx[pid])
                self._log_msg(
                    task_name,
                    f"checkpoint {resource} usage : {checkpoint_usage} MB",
                    logging.DEBUG,
                )
                children_pid_usages = {}
                try:
                    processes = self._get_all_relevant_processes(pid)
                except psutil.NoSuchProcess as err:
                    self._log_with_meta(
                        task_name,
                        f"already finished, {err}",
                        logging.INFO,
                    )
                    continue
                current_usage = get_pid_usage(pid)
                for process in processes[1:]:
                    process_usage = get_pid_usage(process.pid)
                    children_pid_usages[process.pid] = process_usage
                    current_usage += process_usage
                if children_pid_usages:
                    self._log_block_msg(
                        task_name,
                        f"children {resource} usages",
                        self.get_dict_block_msg(children_pid_usages),
                        logging.DEBUG,
                    )
                self._log_msg(
                    task_name,
                    f"current {resource} usage : {current_usage} MB",
                    logging.DEBUG,
                )
                if current_usage == 0:
                    continue
                running_usage = running_pid_usages[pid]
                self._log_msg(
                    task_name,
                    f"running {resource} usage : {running_usage} MB",
                    logging.DEBUG,
                )
                if running_usage > checkpoint_usage:
                    running_pid_counters[pid] = self._refresh_patience
                    self._log_msg(
                        task_name,
                        f"increasing {resource} counter directly to "
                        f"{self._refresh_patience}",
                        logging.DEBUG,
                    )
                elif abs(running_usage - current_usage) <= info["counter_threshold"]:
                    running_pid_counters[pid] += 1
                    self._log_msg(
                        task_name,
                        f"increasing {resource} counter to "
                        f"{running_pid_counters[pid]}",
                        logging.DEBUG,
                    )
                else:
                    running_pid_counters[pid] = 0
                    msg = f"reset {resource} counter"
                    self._log_msg(task_name, msg, logging.DEBUG)
                running_pid_usages[pid] = current_usage
                if running_pid_counters[pid] >= self._refresh_patience:
                    d_usage = current_usage - checkpoint_usage
                    self._log_msg(
                        task_name,
                        f"delta {resource} usage : {d_usage} MB",
                        logging.DEBUG,
                    )
                    checkpoint_pid_usages[pid] = current_usage
                    inference_frees = info["inference_frees"]
                    actual_frees = info["get_available_dict"]()
                    inference_frees[info["pid2resource_id"][pid]] -= d_usage
                    self._log_block_msg(
                        task_name,
                        f"inference {resource} frees after updating delta usage",
                        self.get_dict_block_msg(inference_frees),
                        logging.DEBUG,
                    )
                    self._log_block_msg(
                        task_name,
                        f"actual {resource} frees",
                        self.get_dict_block_msg(actual_frees),
                        logging.DEBUG,
                    )
                    running_pid_counters[pid] = 0
                    msg = f"reset {resource} counter"
                    self._log_msg(task_name, msg, logging.DEBUG)

    def get_process(
        self,
        task_idx: int,
        sleep_method: Callable,
        start: bool,
    ) -> Dict[str, Any]:
        task_name = self._get_task_name(task_idx)
        results: Dict[str, Any] = {
            "__task_name__": task_name,
            "__create_process__": True,
        }
        for resource in self._resources:
            info = self._info_dict[resource]
            local_results = results.setdefault(resource, {})
            preset_usage = info["preset_usages"].setdefault(
                task_idx,
                self.default_usage(resource),
            )
            self._log_with_meta(
                task_name,
                f"preset {resource} usage : {preset_usage} MB; "
                f"minimum {resource} memory needed : {info['minimum_resource']} MB",
            )
            try:
                while True:
                    inference_frees, minimum_resource = map(
                        info.get,
                        ["inference_frees", "minimum_resource"],
                    )
                    self._log_msg(task_name, f"checking {resource}")
                    if len(inference_frees) == 1:
                        tgt_resource_id = next(iter(inference_frees.keys()))
                        free_memory = inference_frees[tgt_resource_id]
                    else:
                        tgt_resource_id, free_memory = sorted(
                            inference_frees.items(), key=lambda kv: kv[1]
                        )[-1]
                    self._log_msg(
                        task_name,
                        f"best choice : {resource} {tgt_resource_id} with "
                        f"{free_memory} MB free memory",
                        logging.DEBUG,
                    )
                    local_results["preset_usage"] = preset_usage
                    local_results["tgt_resource_id"] = tgt_resource_id
                    if preset_usage < free_memory - minimum_resource:
                        self._log_msg(
                            task_name,
                            f"acceptable, {resource} checked",
                            logging.DEBUG,
                        )
                        break
                    if not start:
                        self._log_msg(
                            task_name,
                            "not acceptable, also break out because it's initializing",
                            logging.DEBUG,
                        )
                        results["__create_process__"] = False
                        break
                    self._log_msg(task_name, "not acceptable, waiting", logging.DEBUG)
                    sleep_method()
                if results["__create_process__"]:
                    inference_frees[tgt_resource_id] -= preset_usage
            except KeyboardInterrupt:
                results["__create_process__"] = False
                return results
        return results

    def record_process(
        self,
        task_idx: int,
        process: psutil.Process,
        rs: Dict[str, Any],
    ) -> None:
        pid = process.pid
        task_name = self._get_task_name(task_idx)
        rs_task_name = rs["__task_name__"]
        assert_msg = f"internal error occurred, {task_name} != {rs_task_name}"
        assert task_name == rs_task_name, assert_msg
        task_info, overwritten = None, False
        self.pid2task_idx[pid] = task_idx
        for resource in self._resources:
            local_rs = rs[resource]
            info = self._info_dict[resource]
            pid2resource_id = info["pid2resource_id"]
            checkpoint_pid_usages, running_pid_usages, running_pid_counters = map(
                info.get,
                [
                    "checkpoint_pid_usages",
                    "running_pid_usages",
                    "running_pid_counters",
                ],
            )
            if pid in pid2resource_id:
                self._log_with_meta(
                    task_name,
                    f"pid ({pid}) collision started for {resource}",
                    logging.WARNING,
                )
                if not overwritten:
                    overwritten = True
                    task_info = self._overwritten_task_info.setdefault(task_name, {})
                    task_info["pid"], task_info["task_idx"] = (
                        pid,
                        self.pid2task_idx[pid],
                    )
                task_resource_info = task_info.setdefault(f"{resource}_info", {})
                task_resource_info["ckpt_usage"] = checkpoint_pid_usages[pid]
                task_resource_info["running_usage"] = running_pid_usages[pid]
                task_resource_info["resource_id"] = pid2resource_id[pid]
            running_pid_counters[pid] = 0
            resource_id = pid2resource_id[pid] = local_rs["tgt_resource_id"]
            usage = local_rs["preset_usage"]
            checkpoint_pid_usages[pid] = running_pid_usages[pid] = usage
            self._log_with_meta(
                task_name,
                f"record : using {resource} {resource_id} with {usage} MB "
                f"memory usage (pid : {pid})",
            )

    def handle_finish(
        self,
        task_idx: int,
        process: psutil.Process,
    ) -> Union[str, None]:
        if process is None:
            return
        pid = process.pid
        task_name = self._get_task_name(task_idx)
        self._init_logger(task_name)
        task_info = None
        if task_name not in self._overwritten_task_info:
            pid_task_idx = self.pid2task_idx.pop(pid)
        else:
            task_info = self._overwritten_task_info.pop(task_name)
            recorded_pid, pid_task_idx = map(task_info.get, ["pid", "task_idx"])
            assert_msg = f"internal error occurred ({pid} != {recorded_pid})"
            assert pid == recorded_pid, assert_msg
        msg = "task_idx should be identical with pid_task_idx, internal error occurred"
        assert task_idx == pid_task_idx, msg
        for resource in self._resources:
            resource_info = self._info_dict[resource]
            if task_info is not None:
                task_resource_info = task_info[f"{resource}_info"]
                resource_id = task_resource_info["resource_id"]
                pid_ckpt_usage, pid_running_usage = map(
                    task_resource_info.get, ["ckpt_usage", "running_usage"]
                )
                self._log_with_meta(
                    task_name,
                    f"pid ({pid}) collision ended for {resource}",
                    logging.WARNING,
                )
            else:
                pid_ckpt_usage, pid_running_usage, _ = map(
                    dict.pop,
                    [
                        resource_info["checkpoint_pid_usages"],
                        resource_info["running_pid_usages"],
                        resource_info["running_pid_counters"],
                    ],
                    [pid] * 3,
                )
                resource_id = resource_info["pid2resource_id"].pop(pid)
            pid_usage = pid_running_usage
            if pid_running_usage != pid_ckpt_usage:
                self._log_with_meta(
                    task_name,
                    f"running pid {resource} usage ({pid_running_usage}) != "
                    f"checkpoint pid {resource} usage ({pid_ckpt_usage}), "
                    f"checkpoint {resource} usage will be used to inference "
                    f"{resource} free",
                    logging.WARNING,
                )
                pid_usage = pid_ckpt_usage
            preset_usage = resource_info["preset_usages"][task_idx]
            if pid_running_usage >= preset_usage + resource_info["warning_threshold"]:
                self._log_with_meta(
                    task_name,
                    f"running pid {resource} usage ({pid_running_usage}) exceeded "
                    f"preset cuda {resource} usage ({preset_usage}) so much, "
                    f"it may cause {resource} out of memory",
                    logging.WARNING,
                )
            self._log_with_meta(
                task_name,
                f"finished, releasing {pid_ckpt_usage} MB (inference) memory "
                f"from {resource} {resource_id}",
            )
            inference_frees = resource_info["inference_frees"]
            inference_frees[resource_id] += pid_usage
            for name in [task_name, self._meta_log_name]:
                self._log_block_msg(
                    name,
                    f"inference {resource} frees after releasing",
                    self.get_dict_block_msg(inference_frees),
                    logging.DEBUG,
                )
                self._log_block_msg(
                    name,
                    f"actual {resource} frees",
                    self.get_dict_block_msg(resource_info["get_available_dict"]()),
                    logging.DEBUG,
                )
        return task_name


__all__ = ["PCManager", "GPUManager", "ResourceManager"]



from .core import ParamsGenerator, params_type
from .data_types import DataType, Iterable, Any, Int, Float, Bool, String
from .distributions import DistributionBase, Uniform, Exponential, Choice


__all__ = [
    "ParamsGenerator",
    "params_type",
    "DataType",
    "Iterable",
    "Any",
    "Int",
    "Float",
    "Bool",
    "String",
    "DistributionBase",
    "Uniform",
    "Exponential",
    "Choice",
]



import numpy as np

from typing import *

from ..bases import Optimizer, GradientDescentMixin


@Optimizer.register("adam")
class Adam(Optimizer):
    def __init__(self, lr, **kwargs):
        super().__init__(lr)
        self._eps = kwargs.get("eps", 1e-8)
        self._beta1 = kwargs.get("beta1", 0.9)
        self._beta2 = kwargs.get("beta2", 0.999)

    def step(self, model: GradientDescentMixin, gradient_dict: Dict[str, np.ndarray]):
        for key, value in gradient_dict.items():
            key1, key2 = f"{key}_1", f"{key}_2"
            cache1 = self._caches.setdefault(key1, np.zeros_like(value))
            cache2 = self._caches.setdefault(key2, np.zeros_like(value))
            cache1 = cache1 * self._beta1 + (1.0 - self._beta1) * value
            cache2 = cache2 * self._beta2 + (1.0 - self._beta2) * value ** 2
            self._caches[key1] = cache1
            self._caches[key2] = cache2
            gradient = cache1 / (np.sqrt(cache2 + self._eps))
            self.apply(key, gradient, model)


__all__ = ["Adam"]



import cv2
import math

import numpy as np
import matplotlib.pyplot as plt

from typing import *
from scipy.ndimage import rotate

from ..utils import Reader
from ...misc import check


filters = {
    "roberts": {1: np.array([[1, 0], [0, -1]]), 2: np.array([[0, 1], [-1, 0]])},
    "laplace": {
        1: np.array([[0, 1, 0], [1, -4, 1], [0, 1, 0]]),
        2: np.array([[1, 1, 1], [1, -8, 1], [1, 1, 1]]),
        3: np.array([[2, -1, 2], [-1, -4, -1], [2, -1, 2]]),
        4: np.array([[-1, 2, -1], [2, -4, 2], [-1, 2, -1]]),
    },
    "prewitt": {
        1: np.array([[1, 1, 1], [0, 0, 0], [-1, -1, -1]]),
        2: np.array([[0, 1, 1], [-1, 0, 1], [-1, -1, 0]]),
        3: np.array([[-1, 0, 1], [-1, 0, 1], [-1, 0, 1]]),
        4: np.array([[-1, -1, 0], [-1, 0, 1], [0, 1, 1]]),
        5: np.array([[-1, -1, -1], [0, 0, 0], [1, 1, 1]]),
        6: np.array([[0, -1, -1], [1, 0, -1], [1, 1, 0]]),
        7: np.array([[1, 0, -1], [1, 0, -1], [1, 0, -1]]),
        8: np.array([[1, 1, 0], [1, 0, -1], [0, -1, -1]]),
    },
    "sobel": {
        1: np.array([[1, 2, 1], [0, 0, 0], [-1, -2, -1]]),
        2: np.array([[0, 1, 2], [-1, 0, 1], [-2, -1, 0]]),
        3: np.array([[-1, 0, 1], [-2, 0, 2], [-1, 0, 1]]),
        4: np.array([[-2, -1, 0], [-1, 0, 1], [0, 1, 2]]),
        5: np.array([[-1, -2, -1], [0, 0, 0], [1, 2, 1]]),
        6: np.array([[0, -1, -2], [1, 0, -1], [2, 1, 0]]),
        7: np.array([[1, 0, -1], [2, 0, -2], [1, 0, -1]]),
        8: np.array([[2, 1, 0], [1, 0, -1], [0, -1, -2]]),
    },
    "robinson": {
        1: np.array([[1, 1, 1], [1, -2, 1], [-1, -1, -1]]),
        2: np.array([[1, 1, 1], [-1, -2, 1], [-1, -1, 1]]),
        3: np.array([[-1, 1, 1], [-1, -2, 1], [-1, 1, 1]]),
        4: np.array([[-1, -1, 1], [-1, -2, 1], [1, 1, 1]]),
        5: np.array([[-1, -1, -1], [1, -2, 1], [1, 1, 1]]),
        6: np.array([[1, -1, -1], [1, -2, -1], [1, 1, 1]]),
        7: np.array([[1, 1, -1], [1, -2, -1], [1, 1, -1]]),
        8: np.array([[1, 1, 1], [1, -2, -1], [1, -1, -1]]),
    },
    "kirsch": {
        1: np.array([[3, 3, 3], [3, 0, 3], [-5, -5, -5]]),
        2: np.array([[3, 3, 3], [-5, 0, 3], [-5, -5, 3]]),
        3: np.array([[-5, 3, 3], [-5, 0, 3], [-5, 3, 3]]),
        4: np.array([[-5, -5, 3], [-5, 0, 3], [3, 3, 3]]),
        5: np.array([[-5, -5, -5], [3, 0, 3], [3, 3, 3]]),
        6: np.array([[3, -5, -5], [3, 0, -5], [3, 3, 3]]),
        7: np.array([[3, 3, -5], [3, 0, -5], [3, 3, -5]]),
        8: np.array([[3, 3, 3], [3, 0, -5], [3, -5, -5]]),
    },
}


class Processor:
    def __init__(
        self,
        *,
        rgb: np.ndarray = None,
        bgr: np.ndarray = None,
        img_path: str = None,
    ):
        if rgb is not None or bgr is not None:
            img_array = bgr if bgr is not None else rgb[..., ::-1]
            img_array = img_array.copy()
            gray = img_array.mean(-1).astype(np.uint8)
        else:
            reader = Reader(img_path)
            if not reader.is_valid:
                raise ValueError(f"'{img_path}' is not a valid image")
            img_array = reader.img_array
            gray = reader.to_gray()
        self.img_gray = gray
        self.result = self.img_array = img_array
        self._in_pipeline = False

    def enter_pipeline(self) -> "Processor":
        self.result = self.img_array
        self._in_pipeline = True
        return self

    def exit_pipeline(self) -> "Processor":
        self._in_pipeline = False
        return self

    @property
    def result_is_gray(self) -> bool:
        return len(self.result.shape) == 2

    @property
    def current_array(self) -> np.ndarray:
        return self.result if self._in_pipeline else self.img_array

    @property
    def current_gray(self) -> np.ndarray:
        if not self._in_pipeline:
            return self.img_gray
        if self.result_is_gray:
            return self.result
        return cv2.cvtColor(self.result, cv2.COLOR_RGB2GRAY)

    # smoothen methods

    @check({"ksize": "int"})
    def average(self, *, ksize: int = 3) -> "Processor":
        self.result = cv2.blur(self.current_array, (ksize, ksize))
        return self

    @check({"ksize": "int"})
    def median(self, *, ksize: int = 3) -> "Processor":
        self.result = cv2.medianBlur(self.current_array, ksize)
        return self

    @check({"ksize": ["int", "odd"], "sigma_x": "float", "sigma_y": "float"})
    def gaussian(
        self, *, ksize: int = 3, sigma_x: float = 0.0, sigma_y: float = 0.0
    ) -> "Processor":
        self.result = cv2.GaussianBlur(
            self.current_array, (ksize, ksize), sigma_x, sigmaY=sigma_y
        )
        return self

    # sharpen & edge detection methods

    def _filter(self, name: str, i_type: int, is_edge: bool) -> np.ndarray:
        kernel = filters[name][i_type].copy()
        if not is_edge:
            kernel *= -1
            kernel[1][1] += 1
        return cv2.filter2D(self.current_array, -1, kernel)

    @check({"i_type": ["choices", list(range(1, 3))]})
    def roberts(self, *, i_type: int = 1, is_edge: bool = False):
        self.result = self._filter("roberts", i_type, is_edge)
        return self

    @check({"i_type": ["choices", list(range(1, 5))]})
    def laplace(self, *, i_type: int = 1, is_edge: bool = False):
        self.result = self._filter("laplace", i_type, is_edge)
        return self

    @check({"i_type": ["choices", list(range(1, 9))]})
    def prewitt(self, *, i_type: int = 1, is_edge: bool = False):
        self.result = self._filter("prewitt", i_type, is_edge)
        return self

    @check({"i_type": ["choices", list(range(1, 9))]})
    def sobel(self, *, i_type: int = 1, is_edge: bool = False):
        self.result = self._filter("sobel", i_type, is_edge)
        return self

    @check({"i_type": ["choices", list(range(1, 9))]})
    def robinson(self, *, i_type: int = 1, is_edge: bool = False):
        self.result = self._filter("robinson", i_type, is_edge)
        return self

    @check({"i_type": ["choices", list(range(1, 9))]})
    def kirsch(self, *, i_type: int = 1, is_edge: bool = False):
        self.result = self._filter("kirsch", i_type, is_edge)
        return self

    # corner detection methods

    @check(
        {
            "ksize": ["int", "odd"],
            "k": "float",
            "threshold": "float",
            "block_size": "int",
        }
    )
    def harris(
        self,
        *,
        ksize: int = 3,
        k: float = 0.04,
        threshold: float = 0.01,
        block_size: int = 2,
    ) -> "Processor":
        img = self.current_array.copy()
        gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY).astype(np.float32)
        dst = cv2.cornerHarris(gray, block_size, ksize, k)
        dst = cv2.dilate(dst, None)
        img[dst > threshold * dst.max()] = [0, 0, 255]
        self.result = img
        return self

    # degradation methods

    @check({"mean": "float", "sigma": "float"})
    def gaussian_noise(self, mean: float = 0.0, sigma: float = 51.0) -> "Processor":
        noise = np.random.normal(mean, sigma, self.current_array.shape)
        self.result = np.clip(self.current_array + noise, 0.0, 255.0).astype(np.uint8)
        return self

    @check({"length": "int", "angle": "float"})
    def motion(self, length: int = 8, angle: float = 0.0) -> "Processor":
        h, w, c = self.current_array.shape
        kernel = np.zeros([h, w])
        kernel[
            int(h / 2) : int(h / 2 + 1),
            int(w / 2 - length / 2) : int(w / 2 + length / 2),
        ] = 1
        kernel = rotate(kernel, angle, reshape=False)
        kernel /= kernel.sum()
        self.result = cv2.filter2D(self.current_array, -1, kernel)
        return self

    # thresholding methods

    @check({"max_val": "int", "eps": "float"})
    def global_thresh(
        self,
        max_val: int = 255,
        *,
        eps: float = 1.0,
        return_info: bool = False,
    ) -> Union["Processor", Any]:
        img = self.current_gray
        threshold = img.mean()
        while True:
            mask = img >= threshold
            g1, g2 = img[mask], img[~mask]
            m1, m2 = g1.mean(), g2.mean()
            new_threshold = 0.5 * (m1 + m2)
            if abs(threshold - new_threshold) <= eps:
                threshold = new_threshold
                break
            threshold = new_threshold
        if return_info:
            return g1, g2, m1, m2
        self.result = cv2.threshold(img, threshold, max_val, cv2.THRESH_BINARY)[1]
        return self

    @check(
        {
            "max_val": "int",
            "method": ["choices", ["mean", "gaussian"]],
            "block_size": ["int", "odd"],
            "c": "float",
        }
    )
    def adaptive_thresh(
        self,
        max_val: int = 255,
        *,
        method: str = "gaussian",
        block_size: int = 11,
        c: int = 2,
    ) -> "Processor":
        method = (
            cv2.ADAPTIVE_THRESH_GAUSSIAN_C
            if method == "gaussian"
            else cv2.ADAPTIVE_THRESH_MEAN_C
        )
        self.result = cv2.adaptiveThreshold(
            self.current_gray,
            max_val,
            method,
            cv2.THRESH_BINARY,
            block_size,
            c,
        )
        return self

    @check(
        {
            "max_val": "int",
            "ksize": "int",
            "method": ["choices", ["otsu", "gaussian"]],
            "eps": "float",
        }
    )
    def optimal_thresh(
        self,
        max_val: int = 255,
        *,
        ksize: int = 5,
        method: str = "otsu",
    ):
        if method == "otsu":
            blur = cv2.GaussianBlur(self.current_gray, (ksize, ksize), 0.0)
            self.result = cv2.threshold(
                blur, 0, max_val, cv2.THRESH_BINARY + cv2.THRESH_OTSU
            )[1]
        elif method == "gaussian":
            g1, g2, m1, m2 = self.global_thresh(return_info=True)
            ng1, ng2 = map(len, [g1, g2])
            q, s1 = min(ng1 / ng2, ng2 / ng1), g1.std()
            threshold = 0.5 * (m1 + m2) + s1 ** 2 / (m2 - m1) * math.log(q / (1.0 - q))
            self.result = cv2.threshold(
                self.current_gray, threshold, max_val, cv2.THRESH_BINARY
            )[1]
        return self

    # visualization

    def visualize(self, saving_path: str = None) -> "Processor":
        if saving_path is not None:
            cv2.imwrite(saving_path, self.result)
        else:
            img = self.result if self.result_is_gray else self.result[..., ::-1]
            kwargs = {"cmap": "gray"} if self.result_is_gray else {}
            plt.imshow(img, **kwargs)
            plt.show()
        return self


__all__ = ["Processor"]



from typing import *

number_type = Union[int, float]
generic_number_type = Union[number_type, Any]
nullable_number_type = Union[number_type, None]
bounds_type = Tuple[generic_number_type, generic_number_type]


__all__ = ["number_type", "generic_number_type", "nullable_number_type", "bounds_type"]



import numpy as np

from functools import partial

from ..misc import StrideArray


def naive_rolling_sum(
    array: np.ndarray,
    window: int,
    mean: bool,
    axis: int = -1,
) -> np.ndarray:
    if window > array.shape[axis]:
        raise ValueError("`window` is too large for current array")
    pad_width = [[0, 0] for _ in range(len(array.shape))]
    pad_width[axis][0] = 1
    pad = partial(np.pad, mode="constant", pad_width=pad_width)
    nan_mask = np.isnan(array).astype(np.float32)
    nan_mask = np.swapaxes(pad(nan_mask, constant_values=1.0), axis, -1)
    arr = np.swapaxes(pad(array, constant_values=0.0), axis, -1)

    def _rolling_sum(arr_: np.ndarray) -> np.ndarray:
        cumsum = np.nancumsum(arr_, axis=-1)
        return cumsum[..., window:] - cumsum[..., :-window]

    arr_rolled, nan_rolled = map(_rolling_sum, [arr, nan_mask])
    all_nan_mask = nan_rolled == window
    dtype = arr_rolled.dtype
    arr_rolled = arr_rolled.astype(np.float64)
    if not mean:
        arr_rolled[all_nan_mask] = np.nan
    else:
        nan_rolled[all_nan_mask] = np.nan
        arr_rolled /= window - nan_rolled
    return np.swapaxes(arr_rolled.astype(dtype), axis, -1)


def naive_rolling_min(array: np.ndarray, window: int, axis: int = -1) -> np.ndarray:
    while axis < 0:
        axis += len(array.shape)
    return np.nanmin(StrideArray(array).roll(window, axis=axis), axis + 1)


def naive_rolling_max(array: np.ndarray, window: int, axis: int = -1) -> np.ndarray:
    while axis < 0:
        axis += len(array.shape)
    return np.nanmax(StrideArray(array).roll(window, axis=axis), axis + 1)


def naive_ema(array: np.ndarray, window: int, axis: int = -1) -> np.ndarray:
    shape = list(array.shape)
    while axis < 0:
        axis += len(shape)
    rolled = StrideArray(array).roll(window, axis=axis)
    rolled_axis = axis + 1
    shape.insert(rolled_axis, window)
    new_shape = [1] * len(shape)
    new_shape[rolled_axis] = -1
    ratio = 2.0 / (window + 1.0)
    multipliers = ((1.0 - ratio) ** np.arange(window))[::-1].reshape(new_shape)
    return ratio * np.nansum(rolled * multipliers, rolled_axis)


__all__ = [
    "naive_rolling_sum",
    "naive_rolling_min",
    "naive_rolling_max",
    "naive_ema",
]



from .utils import *


__all__ = [
    "Anneal",
    "Metrics",
    "ScalarEMA",
    "Visualizer",
    "Tracker",
    "Estimator",
    "ModelPattern",
    "EnsemblePattern",
    "Comparer",
    "register_metric",
]



import numpy as np

from typing import *

from ..bases import Optimizer, GradientDescentMixin


@Optimizer.register("sgd")
class SGD(Optimizer):
    def __init__(self, lr, **kwargs):
        super().__init__(lr)
        self._momentum = kwargs.get("momentum", 0.0)
        self._nesterov = kwargs.get("nesterov", False)

    def step(self, model: GradientDescentMixin, gradient_dict: Dict[str, np.ndarray]):
        for key, value in gradient_dict.items():
            if self._momentum <= 0.0:
                gradient = value
            else:
                velocity = self._caches.setdefault(key, np.zeros_like(value))
                velocity = self._momentum * velocity + value
                if self._nesterov:
                    velocity = self._momentum * velocity + value
                gradient = self._caches[key] = velocity
            self.apply(key, gradient, model)


__all__ = ["SGD"]



import numpy
import platform
from Cython.Build import cythonize
from setuptools import setup, find_packages, Extension

VERSION = "0.2.0"

DESCRIPTION = "Some commonly used functions and modules"
with open("README.md") as f:
    LONG_DESCRIPTION = f.read()

INSTALL_REQUIRES = [
    "opencv-python",
    "dill",
    "future",
    "psutil",
    "pillow",
    "pathos",
    "cython>=0.29.12",
    "numpy>=1.16.2",
    "scipy>=1.2.1",
    "scikit-learn>=0.20.3",
    "matplotlib>=3.0.3",
    "mkdocs",
    "mkdocs-material",
    "mkdocs-minify-plugin",
    "Pygments",
    "pymdown-extensions",
]
if platform.system() != "Windows":
    INSTALL_REQUIRES.append("SharedArray")

setup(
    name="carefree-toolkit",
    version=VERSION,
    packages=find_packages(exclude=("tests",)),
    install_requires=INSTALL_REQUIRES,
    ext_modules=cythonize(
        Extension(
            "cftool.c.cython_utils",
            sources=["cftool/c/cython_utils.pyx"],
            language="c",
            include_dirs=[numpy.get_include(), "cftool/c"],
            library_dirs=[],
            libraries=[],
            extra_compile_args=[],
            extra_link_args=[],
        )
    ),
    package_data={"cftool.c": ["cython_utils.pyx"]},
    author="carefree0910",
    author_email="syameimaru_kurumi@pku.edu.cn",
    url="https://github.com/carefree0910/carefree-toolkit",
    download_url=f"https://github.com/carefree0910/carefree-toolkit/archive/v{VERSION}.tar.gz",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    keywords="python numpy data-science",
)
